# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "GraphDelta",
]
from typing import Tuple, Callable
from functools import partial as P

from .. import pyzef
from ._core import *
from . import internals
from .zef_functions import func
from ._ops import *
from .op_structs import ZefOp, LazyValue
from .graph_slice import GraphSlice
from .abstract_raes import Entity, Relation, AttributeEntity, TXNode, Root
from ..pyzef.zefops import SerializedValue
from .logger import log
from .VT import Is, Any, ValueType_
from .internals import instantiate_value_node_imp

from abc import ABC
class ListOrTuple(ABC):
    pass
for t in [list, tuple]:
    ListOrTuple.register(t)


gd_timing = check_env_bool("ZEFDB_GRAPH_DELTA_TIMING")
gd_perform_transaction_debug = check_env_bool("ZEFDB_GRAPH_DELTA_PERFORM_DEBUG")

##############################
# * Description
#----------------------------

# Glossary:
#
# - commands: dictionaries in the constructed GraphDelta which match directly to low-level changes.
# - expressions: high-level things which match to user expectations. There are "complex" expressions and "primitive" expressions.
#

# 1. Entrypoints:
# There are two entrypoints to the code in this file. Either:
#
# receipt = [... list of expressions ...] | transact[g] | run
# or
# unpacked_template = template | g | run
#
# The first is a "GraphDelta" whereas the second is the shorthand syntax.
#
# The shorthand syntax goes through `encode` which will first assign IDs to any
# parts necessary for unpacking the result into the same template structure. It
# then passes the result to `construct_commands`
#
# The `transact[g]` route goes directly to `construct_commands`
#
# 2. `construct_commands`
# Each expressions given by the user is passed to the cmds_for_* functions to
# convert it to either a) an equivalent command, and/or b) additional expressions.
#
# To make this work more easily, `realise_single_node` can take any expression
# which should result in a "single node" (e.g. an ET.Something, or a
# {ET.Something: ...} syntax) and returns a simplified set of expressions to
# realise this, along with the ID to refer to that "single node"
#
# Pretty much the only object that cannot be passed to `realise_single_node` is
# a relation triple. This is handled in `cmds_for_tuple`.
#
# Many expressions passed in are of a LazyValue type. These are consider the
# same as values, and special forms such as assign, set_field or tag are just
# LazyValues that evaluate to themselves.
#
# 3. Flow
#
# (maybe) `encode`
# `construct_commands`:
#     (pre-construct verification)
#     - `obtain_ids` 
#     - `verify_input_el`
#     - `verify_relation_source_target`
#     (construction)
#     - `iteration_step` -> `dispatch_cmds_for` -> `cmds_for_*` -> `realise_single_node`
#     - `verify_and_compact_commands`
# `perform_transaction_commands`
#
# 4. IDs
#
# The IDs must be consistently unique throughout each GraphDelta. A `gen_id`
# object is passed around the functions and produces ids beginning with
# "tmp_id". This is exploited to filter out automatically-generated IDs in the
# receipt.
#
# IDs are how the individual commands are linked to one another. In the end, any
# command that references an ID must come after the ID which creates that item.
# IDs refer to "internal IDs" which do not exist on the graph after the
# transaction finishes. However they are returned in the receipt, allowing user
# code to identify these objects later.



# WARNING!!!
# WARNING!!!
# WARNING!!!
#
# Much of the logic flow in this file is written in an immutable functional
# style. However, for speed many of these functions have been changed to a
# mutating version. This can lead to confusion when debugging.
#
# WARNING!!!
# WARNING!!!
# WARNING!!!


########################################################
# * Entrypoint from user code
#------------------------------------------------------
def dispatch_ror_graph(g, x):
    from . import Effect, FX, Val
    if isinstance(x, LazyValue):
        x = collect(x)
        # Note that this could produce a new LazyValue if the input was an
        # assign_value. This is fine.
    if any(isinstance(x, T) for T in {list, tuple, dict, EntityType, AttributeEntityType, ZefRef, EZefRef, ZefOp, QuantityFloat, QuantityInt, LazyValue, Entity, AttributeEntity, Relation, Val}):
        unpacking_template, commands = encode(x)
        # insert "internal_id" with uid here: the unpacking must get to the RAEs from the receipt
        def insert_id_maybe(cmd: dict):
            if 'origin_rae' in cmd:
                if is_a(cmd['origin_rae'], Delegate):
                    internal_id = cmd['origin_rae']
                else:
                    internal_id = uid(cmd['origin_rae'])
                return {**cmd, 'internal_id': internal_id}
            return cmd

        commands_with_ids = [insert_id_maybe(c) for  c in commands]
        return {
                "type": FX.Graph.Transact,
                "target_graph": g,
                "commands": commands_with_ids,
                "unpacking_template": unpacking_template,
            }
    raise NotImplementedError(f"'x | g' for x of type {type(x)}")


from ..pyzef import main
main.Graph.__ror__ = dispatch_ror_graph




##############################
# * GraphDelta construction
#----------------------------
def construct_commands(elements, generate_id=None) -> list:
    if generate_id is None:
        generate_id = make_generate_id()    # keeps an internal counter        

    # First extract any nested GraphDeltas commands out
    # Note: using native python filtering as the evaluation engine is too slow
    
    # TODO: These were commented out from old GraphDelta behavior. Do we still need to account for it?
    # nested_cmds = elements | filter[is_a[GraphDelta]] | map[lambda x: x.commands] | concat | collect
    # nested_cmds = [x for x in elements if type(x) == GraphDelta] | map[lambda x: x.commands] | concat | collect
    # elements = elements | filter[Not[is_a[GraphDelta]]] | collect
    # elements = [x for x in elements if not type(x) == GraphDelta]

    # The nested cmds need to be readjusted for their auto-generated ids 
    # I want to redo this as pure functional style but will do it with side effects for now
    mapping_dict = {}
    def update_ids(cmd, mapping_dict):
        new_cmd = dict(**cmd)
        if "internal_id" in cmd:
            if is_generated_id(cmd["internal_id"]):
                if cmd["internal_id"] not in mapping_dict:
                    mapping_dict[cmd["internal_id"]] = generate_id()
                new_cmd["internal_id"] = mapping_dict[cmd["internal_id"]]
        for key in ["source", "target"]:
            if key in cmd:
                if cmd[key] in mapping_dict:
                    new_cmd[key] = mapping_dict[cmd[key]]
        return new_cmd

    # nested_cmds = nested_cmds | map[P(update_ids, mapping_dict=mapping_dict)] | collect
    
    # handle any dict if found
    # elements = elements | map[handle_dict | first] | concat | collect
    # Obtain all user IDs
    id_definitions = elements | map[obtain_ids] | reduce[merge_no_overwrite][{}] | collect
    if gd_timing:
        log.debug("Got id_definitions")
    # Check that nothing needs a user ID that wasn't provided
    elements | for_each[verify_input_el[id_definitions]]
    if gd_timing:
        log.debug("Verified input_el")
    elements | for_each[verify_relation_source_target[id_definitions]]
    if gd_timing:
        log.debug("Verified relation_source_target")

    state_initial = {
        'user_expressions': tuple(elements),
        'commands': [],
    }

    def make_debug_output():
        num_done = 0
        next_print = now()+5*seconds
        def debug_output(x):
            nonlocal num_done, next_print
            num_done += 1
            if now() > next_print:
                log.debug("Construct: Up to", num_done=num_done, num_expr=len(x['user_expressions']), num_cmd=len(x['commands']))
                next_print = now() + 5*seconds
        return debug_output
        
    # iterate until all user expressions and generated expressions have become commands
    state_final = (state_initial
                   | iterate[iteration_step[generate_id]]
                   | (map[tap[make_debug_output()]] if gd_timing else identity)
                   | take_until[lambda s: len(s['user_expressions'])==0]
                   # | map[tap[print]]
                   | last
                   | collect
                   )
    return verify_and_compact_commands(state_final['commands'])



##############################
# * Validation
#----------------------------

@func
def obtain_ids(x) -> dict:
    # Return all user IDs for this part of the GraphDelta input.

    # The output of this func (and anytime it is used recursively) must go
    # through merge_no_overwrite to avoid ID clashes.

    ids = {}
    if isinstance(x, LazyValue):
        ids = obtain_ids(x.initial_val)
    elif is_a(x, Delegate):
        ids = {x: x}

    # this must be a triple, defining a relation
    elif isinstance(x, ListOrTuple):
        ids = x | map[obtain_ids] | reduce[merge_no_overwrite][{}] | collect

    elif isinstance(x, dict):
        entity, sub_d = list(x.items()) | single | collect # Will throw if there isn't a single key in the dict

        ids = obtain_ids(entity)
        for k,v in sub_d.items():
            ids = merge_no_overwrite(ids, obtain_ids(k))
            ids = merge_no_overwrite(ids, obtain_ids(v))

    elif type(x) in [Entity, AttributeEntity, Relation]:
        ids = {uid(x): x}
        if type(x) == Relation:
            if not is_a(x.d["type"][0], RT):
                ids = merge_no_overwrite(ids, obtain_ids(source(x)))
            if not is_a(x.d["type"][2], RT):
                ids = merge_no_overwrite(ids, obtain_ids(target(x)))

    elif type(x) in [ZefRef, EZefRef]:
        ids = {origin_uid(x): x}

    elif type(x) == TaggedVal:
        if x.iid is not None:
            ids = {x.iid: x.primitive}


    # This is an extra step on top of the previous checks
    if type(x) in [Entity, AttributeEntity, Relation, EntityType,
                    AttributeEntityType, RelationType, ZefRef, EZefRef]:
        # Need the lazy value for the RT possibility
        a_id = get_absorbed_id(LazyValue(x))
        if a_id is not None:
            ids = merge_no_overwrite(ids, {a_id: x})
    return ids


@func    
def verify_input_el(x, id_definitions, allow_rt=False, allow_scalar=False):
    from . import Val
    # Check each input item to a GraphDelta to validate its form. Uses
    # id_definitions to assert that references to IDs are to things that are
    # defined in the GraphDelta.

    if isinstance(x, LazyValue):
        # This is checking for an (x <= value) format
        if is_a(x.el_ops, assign):
            return
        elif is_a(x.el_ops, terminate):
            return
        elif is_a(x.el_ops, tag):
            return
        elif is_a(x.el_ops, fill_or_attach):
            return
        elif is_a(x.el_ops, set_field):
            return
        raise Exception(f"A LazyValue must have come from (x | assign), (x | terminate), (x | fill_or_attach) or (x | tag) only. Got {x}")

    # Note: just is_a(x, Z) will also mean ZefRefs will be hit
    elif is_a(x, ZefOp) and is_a(x, Z):
        some_id = get_curried_arg(x, 0)

        if some_id not in id_definitions:
            raise KeyError(f"The id '{some_id}' used in Z is not an internal id defined in the GraphDelta init list")
        return

    elif isinstance(x, ZefRef) or isinstance(x, EZefRef):
        return

    elif isinstance(x, dict):
        entity, sub_d = list(x.items()) | single | collect # Will throw if there isn't a single key in the dict
        verify_input_el(entity, id_definitions, False, False)

        for k,v in sub_d.items():
            verify_input_el(k, id_definitions, True, False)
            if isinstance(v, list):
                [verify_input_el(e, id_definitions, False, True) for e in v]
            else:
                verify_input_el(v, id_definitions, False, True)

        return

    elif isinstance(x, ListOrTuple):                         
        if is_valid_relation_template(x):
            if len(x) == 3:
                verify_input_el(x[0], id_definitions, False, True)
                verify_input_el(x[1], id_definitions, True, False)
                verify_input_el(x[2], id_definitions, False, True)
                return
            else:
                return
        for item in x:
            verify_input_el(item, id_definitions, False, allow_scalar)
        return
    elif isinstance(x, RelationType):
        if allow_rt:
            return
        else:
            raise ValueError(f"Bare RTs without source or target cannot be initialized. You tried to create a {x}.")                    
    elif type(x) in scalar_types:
        if not allow_scalar:
            raise Exception("Direct values are not allowed at the top level of a GraphDelta, as this is likely to indicate a typo, e.g. (ET.Machine, ET.ShouldBeRTNotET, 'name'). Please use explicit value assignment via (AET.String <= 'name') if you really want this behaviour.")
        return
    elif is_valid_single_node(x):
        return

    elif isinstance(x, ZefOp):
        if len(x) > 1:
            # This could be something like Z["asdf"] | assign[5]
            if LazyValue(x) | first | is_a[Z] | collect:
                return
        if length(LazyValue(x) | absorbed) != 1:
            raise Exception(f"ZefOp has more than one op inside of it. {x}")
        raise Exception(f"Not allowing ZefOps except for those starting with Z. Got {x}")

    elif is_a(x, Delegate):
        return

    elif type(x) in [Entity, AttributeEntity, Relation]:
        return

    elif type(x) in [Val, TaggedVal]:
        return

    else:
        raise ValueError(f"Unexpected type passed to init list of GraphDelta: {x} of type {type(x)}")

@func    
def verify_relation_source_target(x, id_definitions):
    if type(x) == Relation:
        assert all(u in id_definitions for u in x.d["uids"]), "A Relation doesn't have its corresponding source or target included in the GraphDelta. This is likely because you have an abstract Relation with another Relation as its source/target. These must be included explicitly into the GraphDelta."




##################################################
# * Expr -> cmd conversion
#------------------------------------------------

@func
def iteration_step(state: dict, gen_id)->dict:
    """[summary]

    Args:
        state (dict): of form 
        {
            'user_expressions': elements,
            'commands': (),
        }

    Returns:
        dict: [description]
    """

    # Note: we used to keep track of ids seen here to not merge the same thing
    # twice. However, the user can potentially give two different internal ids
    # to the same merged object, so we can't follow that logic anymore. Instead,
    # the nearly-identical duplicate merges will be contracted later on in the
    # compact commands function.

    exprs = state['user_expressions']
    if len(exprs) == 0:
        return state
    # expr = exprs[0]

    # WARNING: this mutates but is necessary for speed
    if not isinstance(exprs, list):
        exprs = list(exprs)
    expr = exprs.pop()

    new_exprs,new_cmds = dispatch_cmds_for(expr, gen_id)

    # Here are three ways to accomplish the same thing, although the last
    # (and fastest) mutates.

    # return {
    #         'user_expressions': (*exprs[1:], *new_exprs),
    #         'commands': (*state['commands'], *new_cmds),                 # for now this will be a dict to allow quick lookup: the key is the uid / internal id
    #         'ids_present': {*state['ids_present'], *new_ids}
    #     }

    # return {
    #         'user_expressions': exprs[1:] + list(new_exprs),
    #         'commands': state['commands'] + list(new_cmds),
    #         'ids_present': state['ids_present'] | set(new_ids)
    #     }


    exprs.extend(new_exprs)
    cmds = state["commands"]
    cmds.extend(new_cmds)
    return {
            'user_expressions': exprs,
            'commands': cmds,
    }

def dispatch_cmds_for(expr, gen_id):
    from . import Val
    if isinstance(expr, LazyValue):
        expr = collect(expr)

    # If we have a lazy value the second time around, then this must be an actual action to perform.
    if isinstance(expr, LazyValue):
        if is_a(expr.el_ops, assign):
            return cmds_for_lv_assign(expr, gen_id)
        elif is_a(expr.el_ops, terminate):
            return cmds_for_lv_terminate(expr)
        elif is_a(expr.el_ops, set_field):
            return cmds_for_lv_set_field(expr, gen_id)
        elif is_a(expr.el_ops, fill_or_attach):
            return cmds_for_complex_expr(expr, gen_id)
        elif is_a(expr.el_ops, tag):
            return cmds_for_lv_tag(expr, gen_id)
        else:
            raise Exception("LazyValue obtained which is not an assign or terminate")

    elif isinstance(expr, ZefOp):
        # If we have a chain beginning with a Z, convert to LazyValue
        if LazyValue(expr) | first | is_a[Z] | collect:
            return cmds_for_initial_Z(expr)

        raise RuntimeError(f'We should not have landed here, with expr={expr}')
    if is_a(expr, Delegate):
        return cmds_for_delegate(expr)

    d_dispatch = {
        EntityType: cmds_for_instantiable,
        AttributeEntityType: cmds_for_instantiable,

        ZefRef: cmds_for_mergable,
        EZefRef: cmds_for_mergable,
        Entity: cmds_for_mergable,
        AttributeEntity: cmds_for_mergable,
        Relation: cmds_for_mergable,
        TXNode: cmds_for_mergable,
        Root: cmds_for_mergable,

        tuple: P(cmds_for_tuple, gen_id=gen_id),
        list: P(cmds_for_tuple, gen_id=gen_id),

        dict: P(cmds_for_complex_expr, gen_id=gen_id),

        Val: cmds_for_value_node,
        TaggedVal: cmds_for_value_node,
    }
    if type(expr) not in d_dispatch:
        raise TypeError(f"transform_to_commands was called for type {type(expr)} value={expr}, but no handler in dispatch dict")            
    func = d_dispatch[type(expr)]

    return func(expr)



##################################################
# * Handlers for expr->cmd
#------------------------------------------------

def cmds_for_complex_expr(x, gen_id):
    iid,exprs = realise_single_node(x, gen_id)
    return exprs, []

def cmds_for_initial_Z(expr):
    assert LazyValue(expr) | first | And[is_a[ZefOp]][is_a[Z]] | collect

    expr = LazyValue(LazyValue(expr) | first | collect) | (LazyValue(expr) | skip[1] | to_pipeline | collect)
    return (expr,), ()

def cmds_for_value_node(x):
    from . import Val
    if type(x) == Val:
        x = TaggedVal(x.arg, None)

    cmd = {'cmd': 'instantiate_value_node',
           'value': x.primitive}
    
    if x.iid is not None:
        cmd['internal_id'] = x.iid
        
    return (), [cmd]


def cmds_for_instantiable(x):
    a_id = get_absorbed_id(x)
    raet = x | without_absorbed | collect
    cmd = {'cmd': 'instantiate', 'rae_type': raet}
    if a_id is not None:
        cmd['internal_id'] =  a_id
        
    return (), [cmd]

def cmds_for_mergable(x):
    origin = origin_rae(x)

    cmd = {"cmd": "merge",
           "origin_rae": origin,
           "internal_ids": []}

    a_id = get_absorbed_id(x)
    if a_id is not None:
        cmd['internal_ids'] += [a_id]

    if is_a(x, ZefRef) or is_a(x, EZefRef):
        if BT(x) in {BT.ENTITY_NODE, BT.TX_EVENT_NODE, BT.ROOT_NODE}:
            return (), [cmd]

        elif BT(x) == BT.ATTRIBUTE_ENTITY_NODE:
            cmds = [cmd]
            if isinstance(x, ZefRef):
                val = x | value | collect
                if val is not None:
                    cmds += [{
                        'cmd': 'assign', 
                        'value': val,
                        'internal_id': uid(origin),
                        'explicit': False,
                }]

            return (), cmds

        elif BT(x) == BT.RELATION_EDGE:
            return (
                (x | source | to_ezefref | collect, x | target | to_ezefref | collect),
                [cmd],
            )
        else:
            raise NotImplementedError(f"Unknown ZefRef type for merging: {BT(x)}")
    elif is_a(x, Entity) or is_a(x, AttributeEntity):
        return (), [cmd]
    elif is_a(x, Relation):
        maybe_src_trg = []
        if not is_a(x.d["type"][0], RT):
            maybe_src_trg.append(source(x))
        if not is_a(x.d["type"][2], RT):
            maybe_src_trg.append(target(x))

        return (
            tuple(maybe_src_trg),
            [cmd],
        )
    else:
        raise NotImplementedError(f"Unknown type for mergable: {type(x)}")

def cmds_for_lv_tag(x, gen_id):
    assert isinstance(x, LazyValue)

    assert len(x | peel | collect) == 2
    target,op = x | peel | collect

    assert is_a(op, tag)
    tag_s = LazyValue(op) | absorbed | single | collect
    iid,exprs = realise_single_node(target, gen_id)
    cmd = {'cmd': 'tag',
           'tag_name': tag_s,
           'internal_id': iid}
    return exprs, [cmd]
    

def cmds_for_lv_assign(x, gen_id: Callable):
    from . import Val
    assert isinstance(x, LazyValue)

    assert len(x | peel | collect) == 2
    target,op = x | peel | collect

    assert is_a(op, assign)
        
    iid,exprs = realise_single_node(target, gen_id)

    cmd = {'cmd': 'assign', 'explicit': True, "internal_id": iid}

    val = LazyValue(op) | absorbed | single | collect
    if type(val) == Val:
        cmd['value'] = TaggedVal(val.arg, None).primitive
    else:
        cmd['value'] = val

    return exprs, [cmd]

def cmds_for_lv_terminate(x):
    assert isinstance(x, LazyValue)

    assert len(x | peel | collect) == 2
    target,op = x | peel | collect

    assert is_a(op, terminate)
    cmd = {
        'cmd': 'terminate', 
        'origin_rae': origin_rae(target)
        }
    a_id = get_absorbed_id(LazyValue(op))
    if a_id is not None:
        cmd['internal_id'] = a_id

    return (), [cmd]

def cmds_for_lv_set_field(x, gen_id):
    assert isinstance(x, LazyValue)

    assert len(x | peel | collect) == 2
    source,op = x | peel | collect

    # assert is_a(source, ZefOp) and is_a(source, Z), f"fill_or_attach reached cmd creation with an incorrect input type: {source}"
    assert is_a(op, set_field)
    rt,assignment,incoming = LazyValue(op) | absorbed | collect
    assert type(rt) == RelationType

    iid,exprs = realise_single_node(source, gen_id)

    cmd = {
        'cmd': 'set_field', 
        'source_id': iid,
        'rt': LazyValue(rt) | without_absorbed | collect,
        'incoming': incoming,
    }

    if type(assignment) in scalar_types:
        cmd['value'] = assignment 
    else:
        target_iid,target_exprs = realise_single_node(assignment, gen_id)
        cmd['target_id'] = target_iid
        exprs.extend(target_exprs)

    if LazyValue(rt) | absorbed | collect != ():
        cmd['internal_id'] = LazyValue(rt) | absorbed | single | collect

    return exprs, [cmd]

def cmds_for_delegate(x):                            
    internal_ids = []
    a_id = get_absorbed_id(x)
    if a_id is not None:
        internal_ids += [a_id]

    return (), [{
        'cmd': 'merge', 
        'origin_rae': x,
        'internal_ids': internal_ids,
        }]



def cmds_for_tuple(x: tuple, gen_id: Callable):
    new_exprs = []
    new_cmds = []

    if not is_valid_relation_template(x):
        raise Exception("Not allowing list/tuple nested in GraphDelta unless it belongs to a relation")

    if len(x) == 3:
        if isinstance(x[0], ListOrTuple) and isinstance(x[2], ListOrTuple):
            # Case 4 of is_valid_relation_template
            new_exprs += [(source, x[1], target) for source in x[0] for target in x[2]]
        elif isinstance(x[0], ListOrTuple):
            # Case 3 of is_valid_relation_template
            iid,exprs = realise_single_node(x[2], gen_id)
            new_exprs += exprs
            new_exprs += [(source, x[1], Z[iid]) for source in x[0]]
        elif isinstance(x[2], list):
            # Case 2 of is_valid_relation_template
            iid,exprs = realise_single_node(x[0], gen_id)
            new_exprs += exprs
            new_exprs += [(Z[iid], x[1], target) for target in x[2]]
        else:
            # Case 1 of is_valid_relation_template
            # it's a plain relation triple
            src_id, src_exprs = realise_single_node(x[0], gen_id)
            trg_id, trg_exprs = realise_single_node(x[2], gen_id)

            new_exprs += src_exprs
            new_exprs += trg_exprs

            # This is the only time the relation is created
            rel = x[1]
            if isinstance(rel, RelationType):
                a_id = get_absorbed_id(rel)
                raet = LazyValue(rel) | without_absorbed | collect
                cmd1 = {'cmd': 'instantiate', 'rae_type': raet}
                if a_id is not None:
                    cmd1['internal_id'] =  a_id
            elif isinstance(rel, ZefOp):
                cmd1 = {'cmd': 'instantiate',
                        'rae_type': rel.el_ops[0][1][0],
                        'internal_id': rel.el_ops[0][1][1]}
            cmd1['source'] = src_id
            cmd1['target'] = trg_id

            new_cmds += [cmd1]

    elif len(x) == 2:
        # Case 5 of is_valid_relation_template
        iid,exprs = realise_single_node(x[0], gen_id)
        new_exprs += exprs
        new_exprs += [(Z[iid], item[0], item[1]) for item in x[1]]
    else:
        raise Exception("Shouldn't get here")

    return new_exprs, new_cmds





##############################################################
# ** Complex->primitve expr utils
#------------------------------------------------------------

def is_valid_single_node(x):
    if type(x) in scalar_types:
        return True
    if isinstance(x, ZefRef) or isinstance(x, EZefRef):
        return True
    if is_a(x, Z):
        return True
    if isinstance(x, EntityType):
        return True
    if isinstance(x, RelationType):
        return True
    if isinstance(x, AttributeEntityType):
        return True
    if isinstance(x, Delegate):
        return True
    return False

def is_valid_relation_template(x):
    # Check for the relation templates:
    # Either:
    # 1. (a, RT.x, b)
    # 2. (a, RT.x, [b,c,d,e,...])
    # 3. ([a,b,c,...], RT.x, d)
    # 4. ([a,b,c,...], RT.x, [d,e,f,...])
    # 5. (a, [(RT.x, b), (RT.y, c)])
    # Note: tuple/list can be used interchangably
    # Note: 5. deliberately does not have its symmetric counterpart. This could be introduced later.
    if any(isinstance(item, RelationType) for item in x):
        # Cases 1-4 above
        if len(x) != 3:
            raise Exception(f"A list has an RT but isn't 3 elements long: {x}")
        if not isinstance(x[1], RelationType):
            raise Exception(f"A list has an RT but it isn't in the second position: {x}")
        # Note: if there are any lists involved, we cannot have a given ID as it
        # would have to be given to multiple instantiated relations.
        if not is_a(x[1], RT) and (isinstance(x[0], ListOrTuple) or isinstance(x[2], ListOrTuple)):
            raise Exception("An RT with an internal name is not allowed with multiple sources or targets")
        return True
    elif len(x) == 2 and isinstance(x[1], ListOrTuple):
        return all(isinstance(item, ListOrTuple) and len(item) == 2 and isinstance(item[0], RelationType) and is_valid_single_node(item[1]) for item in x[1])
    return False


@func
def realise_single_node(x, gen_id):
    from . import Val
    # Take something that should refer to a single node, i.e. a RAE or a scalar
    # to be turned into a RAE, or a reference to a RAE, and return a version
    # with an explicit ID and the ID itself.
    #
    # For example
    #
    # realise_single_node(ET.Machine["a"])
    # will return
    # ("a", ET.Machine["a"])
    #
    # realise_single_node(ET.Machine)
    # might return
    # ("12323425346", ET.Machine["12323425346"])


    #     5 different cases we need to deal with:
    # 4) A ZefRef / EZefRef to an existing RAE is specified. Generate the merge command if it is not present yet
    # 5) Z['my_temp_id4']       The temp id may not exist yet (could come from another edge creation). Make sure it is checked to exist at the end
    # 3) e.g. if a plain value was specified, create an AE and assign the value
    # 1) e.g. if ET.Dog is specified as the source: definitely create it
    # 2) e.g. if ET.Dog['Rufus'] was specified, that will be there as a dict. It is the command to create it and register the temp id

    # First case of removing lazy values
    if isinstance(x, LazyValue):
        x = collect(x)

    # Now this is a check for whether we are an assign
    if isinstance(x, LazyValue):
        target,op = x | peel | collect
        if is_a(op, terminate) or is_a(op, tag):
            iid,exprs = realise_single_node(target, gen_id)
            exprs = [x]
        elif is_a(op, assign):
            iid,exprs = realise_single_node(target, gen_id)
            exprs = exprs + [LazyValue(Z[iid]) | op]
        elif is_a(op, fill_or_attach):
            # fill_or_attach behaviour is now basically set_field except when the target is not a value
            iid,exprs = realise_single_node(target, gen_id)
            rt,assignment = LazyValue(op) | absorbed | collect
            if type(assignment) in scalar_types:
                exprs = exprs + [LazyValue(Z[iid]) | set_field[rt][assignment]]
            else:
                exprs = exprs + [(Z[iid], rt, assignment)]
        elif is_a(op, set_field):
            iid,exprs = realise_single_node(target, gen_id)
            exprs = exprs + [LazyValue(Z[iid]) | op]
        else:
            raise Exception(f"Don't understand LazyValue type: {op}")
    elif isinstance(x, EntityType) or isinstance(x, AttributeEntityType):
        a_id = get_absorbed_id(x)
        if a_id is None:
            iid = gen_id()
            exprs = [x[iid]]
        else:
            iid = a_id
            exprs = [x]
    elif isinstance(x, ZefRef) or isinstance(x, EZefRef):
        if internals.is_delegate(x):
            d = to_delegate(x)
            exprs = [d]
            iid = d
        else:
            exprs = [x]
            iid = origin_uid(x)
    elif type(x) in [Entity, AttributeEntity, Relation, TXNode, Root]:
        exprs = [x]
        iid = origin_uid(x)
    elif type(x) == Val:
        iid = gen_id()
        exprs = [TaggedVal(x.arg, iid)]
    elif type(x) in shorthand_scalar_types:
        iid = gen_id()
        aet = map_scalar_to_aet_type[type(x)](x)
        exprs = [aet[iid], LazyValue(Z[iid]) | assign[x]]
    elif type(x) in scalar_types:
        raise Exception("A value of type {type(x)} is not allowed to be given in a GraphDelta in the shorthand syntax as it is ambiguous. You might want to explicitly create an AET and assign, or a value node, or a custom AET.")
    elif isinstance(x, ZefOp):
        if len(x) == 1:
            if is_a(x, Z):
                iid = LazyValue(x) | peel | first | second | first | collect
                # No expr to perform
                exprs = []
            else:
                raise NotImplementedError(f"Can't pass zefops to GraphDelta: for {x}")
        else:
            ops = LazyValue(x) | peel | collect
            first_op = ops[0]
            rest = ops[1:]
            if is_a(first_op, Z):
                new_op = LazyValue(first_op) | to_pipeline(rest)
                iid,exprs = realise_single_node(new_op, gen_id)
                print(iid)
            else:
                raise NotImplementedError(f"Can't pass zefops to GraphDelta: for {x}")
    elif is_a(x, Delegate):
        iid = x
        exprs = [x]
    elif type(x) == dict:
        entity, sub_d = list(x.items()) | single | collect # Will throw if there isn't a single key in the dict
        iid, entity_exprs = realise_single_node(entity, gen_id)

        exprs = entity_exprs
        for k,v in sub_d.items():
            if isinstance(v, list):
                # for e in v:
                #     assert type(e) not in scalar_types, "Can't have multiple scalars attached in a dictionary"
                #     exprs.append((Z[iid], k, e))
                raise Exception("Not allowed to use lists inside of a dictionary syntax anymore")
            else:
                exprs.append(Z[iid] | set_field[k][v])
    else:
        raise TypeError(f'in GraphDelta encode step: for type(x)={type(x)}')

    return iid, exprs



##############################
# * Ordering commands
#----------------------------

def verify_and_compact_commands(cmds: tuple):                
    aes_with_explicit_assigns = (cmds
                                 | filter[lambda d: d["cmd"] == "assign"]
                                 | filter[lambda d: d["explicit"]]
                                 | map[lambda d: d["internal_id"]]
                                 | collect
                                 )
    def is_unnecessary_automatic_merge_assign(d):
        return (d["cmd"] == "assign"
                and not d["explicit"]
                and d["internal_id"] in aes_with_explicit_assigns)

    # make sure if multiple assignment commands exist for the same AE, that their values agree
    cmds = (cmds | group_by[get['cmd']]
            | map[match[
                        (Is[first | equals["assign"]],   second | filter[Not[is_unnecessary_automatic_merge_assign]] | validate_and_compress_unique_assignment),
                # TODO: Validate fill_or_attach assignments - but note that
                # these could also conflict with regular assignments but it's
                # hard to tell that without playing out the assignments first.
                        (Is[first | equals["merge"]],          second | combine_internal_ids_for_merges),
                        (Is[first | equals["terminate"]],      second | combine_terminates), 
                        (Is[first | equals["instantiate"]],    second | combine_instantiates), 
                        (Any, second),                                 # Just pack things back in for other cmd types
            ]]
            | concat
            | collect)

    sorted_cmds = tuple(cmds 
                        | sort[command_ordering_by_type] 
                        | collect
                        )

    def make_debug_output():
        num_done = 0
        next_print = now()+5*seconds
        def debug_output(x):
            nonlocal num_done, next_print
            num_done += 1
            if now() > next_print:
                log.debug("Compacting:", num_done=num_done, num_input=len(x['state']['input']), num_output=len(x['state']['output']))
                next_print = now() + 5*seconds
        return debug_output

    state_initial = {
        'input': sorted_cmds,
        'output': (),
        'known_ids': set(),
    }
    state_final = (
        {"state": state_initial, "num_changed": -1}
        | iterate[resolve_dag_ordering_step] 
        | (map[tap[make_debug_output()]] if gd_timing else identity)
        | take_until[lambda s: s["num_changed"] == 0]
        # | map[tap[print]]
        | last
        | get["state"]
        | collect
    )
    # print("Done state_final", now())
    if len(state_final['input']) > 0:
        import json
        print(json.dumps(state_final, indent=4, default=repr))
        raise NotImplementedError("Error constructing GraphDelta: instantiation order iteration did not converge. Probably there is a circular dependency in the required imperative instantiation order between inter-dependent relations. This is a valid GraphDelta in principle, but currently not implemented in zefDB.")
    ordered_cmds = state_final['output']
    return ordered_cmds

@func
def validate_and_compress_unique_assignment(cmds):
    @func
    def check_list_is_distinct(cmds):
        values = cmds | map[get["value"]] | distinct | collect
        if length(values) == 1:
            return cmds[0]
        raise ValueError(f'There may be at most one assignment commands for each AE. There were multiple for assignment to {get_id(cmds[0])!r} with values {values}')

    return (cmds
            | group_by[get_id]
            | map[second | check_list_is_distinct])
            

@func
def combine_internal_ids_for_merges(cmds):
    def combine_ids(x, y):
        assert x["origin_rae"] == y["origin_rae"]
        return {**x,
                "internal_ids": x["internal_ids"] + y["internal_ids"]}
    
    return (cmds
              # Convert each item to a dictionary using the origin_rae as key
            | map[lambda x: {x["origin_rae"]: x}]
              | reduce[merge_with[combine_ids]][{}]
              # Now we should have one big dict with each command
            | values)

@func
def combine_terminates(cmds):
    @func
    def check_list_is_distinct(cmds):
        # cmds = cmds | distinct | collect
        # if length(cmds) == 1:
        #     return cmds[0]
        if not cmds | map[equals[first(cmds)]] | all | collect:
            ids = cmds | map[get["internal_id"][None]] | collect
            raise ValueError(f'There may be at most one id for a terminated ZefRef. There were multiple for assignment to {cmds[0]["origin_rae"]!r} with ids {ids}')
        return cmds[0]

    return (cmds
            | group_by[get["origin_rae"]]
            | map[second | check_list_is_distinct])

@func
def combine_instantiates(cmds):
    @func
    def check_list_is_distinct(arg):
        iid,cmds = arg
        if iid is None:
            return cmds
        if not cmds | map[equals[first(cmds)]] | all | collect:
            raets = cmds | map[get["rae_type"]] | collect
            raise ValueError(f'There may be at most one RAET for an instantiated ZefRef. Found: {raets!r} for id {cmds[0]["internal_id"]}')
        return [cmds[0]]

    return (cmds
            | group_by[get["internal_id"][None]]
            | map[check_list_is_distinct]
            | concat)



def command_ordering_by_type(d_raes: dict) -> int:
    """we want some standardized order of the output to simplify value-based
        comparisons and other operations for graph deltas"""
        
    if d_raes['cmd'] == 'merge':
        if isinstance(d_raes['origin_rae'], Relation): return 0.5
        else: return 0
    if d_raes['cmd'] == 'instantiate_value_node':
        # Comes before AET in case there is an assignment there
        return 1.5
    if d_raes['cmd'] == 'instantiate':
        if isinstance(d_raes['rae_type'], EntityType): return 1
        if isinstance(d_raes['rae_type'], AttributeEntityType): return 2
        if isinstance(d_raes['rae_type'], RelationType): return 3
        return 4                                            # there may be {'cmd': 'instantiate', 'rae_type': AET.Bool}
    if d_raes['cmd'] == 'assign': return 5
    if d_raes['cmd'] == 'set_field': return 5.5
    if d_raes['cmd'] == 'terminate' and isinstance(d_raes['origin_rae'], Relation): return 6
    if d_raes['cmd'] == 'terminate' and isinstance(d_raes['origin_rae'], Entity): return 7
    if d_raes['cmd'] == 'terminate' and isinstance(d_raes['origin_rae'], AttributeEntity): return 8
    if d_raes['cmd'] == 'terminate' and is_a(d_raes['origin_rae'], Delegate): return 9
    if d_raes['cmd'] == 'tag': return 10
    else: raise NotImplementedError(f"In Sort fct for {d_raes}")

def get_id(cmd):
    if "internal_ids" in cmd:
        raise Exception("Not allowed to ask for id of command that could have multiple ids")
    elif 'internal_id' in cmd:
        return cmd['internal_id']
    elif 'origin_rae' in cmd:
        if is_a(cmd['origin_rae'], Delegate):
            return cmd['origin_rae']
        return uid(cmd['origin_rae'])
    else:
        return None

def get_ids(cmd):
    ids = []
    if cmd['cmd'] == 'merge':
        ids += cmd["internal_ids"]
        if is_a(cmd['origin_rae'], Delegate):
            ids += [cmd['origin_rae']]
        else:
            ids += [uid(cmd['origin_rae'])]
    else:
        this_id = get_id(cmd)
        if this_id is not None:
            ids += [this_id]
    return ids

def resolve_dag_ordering_step(arg: dict)->dict:
    # arg is "state" + "num_changed"
    state = arg["state"]
    ids = state['known_ids']        
    input = state['input']
    
    def can_be_executed(cmd):        
        if cmd['cmd'] == 'instantiate':
            # If we are creating an RT, wait until both source/target exist
            return not isinstance(cmd['rae_type'], RelationType) or (cmd['source'] in ids and cmd['target'] in ids)
        if cmd['cmd'] == 'instantiate_value_node':
            return True
        if cmd['cmd'] == 'merge':
            # If the merge is of a relation, we need both source and target to exist already
            return not isinstance(cmd['origin_rae'], Relation) or (cmd['origin_rae'].d["uids"][0] in ids and cmd['origin_rae'].d["uids"][2] in ids)
        if cmd['cmd'] == 'terminate':
            # Don't terminate if there an upcoming operation that will create this item
            return all(get_id(cmd) not in get_ids(other) for other in input if other['cmd'] != 'terminate')
        if cmd['cmd'] == 'assign':
            # The object to be assigned needs to exist.
            return get_id(cmd) in ids
        if cmd['cmd'] == 'set_field':
            # This is where things get tricky - the behaviour of set_field
            # can change if there is another command that creates a relation of
            # the same type.
            #
            # TODO: this properly! I can see issues with assignments to the same
            # object via different IDs causing massive headaches here. Need to
            # consider aliasing and also consider multiple set_field commands.
            #
            # For now, just make sure the source is alive and run with that.
            return cmd['source_id'] in ids
        if cmd['cmd'] == 'tag':
            return get_id(cmd) in ids
        raise Exception(f"Don't know how to decide DAG ordering for command {cmd['cmd']}")
    (_,can), (_,cannot) = state['input'] | group_by[can_be_executed][[True, False]] | collect
    return {
        "state": {
            'input': tuple(cannot),
            'output': (*state['output'], *can),
            'known_ids': {*state['known_ids'], *(can | map[get_ids] | concat)},
        },
        "num_changed": len(can) > 0
    }
    
        

########################################################
# * Shorthand syntax handlers
#------------------------------------------------------



def encode(xx):
    """
    This function is invoked if one writes "[ET.Dog, (ET.Person, ET.Aardvark)] | g"
    It returns a GraphDelta with ids assigned, as well as a schema
    with the same structure as the nested input arrays, but the ids as placeholders

    Args:
        xx ([type]): [description]

    Returns:
        [type]: Tuple, 
    """
    gd_exprs = []
    
    gen_id = make_generate_id()

    # - encode works recursively
    # - step handles each recursive step
    #
    # logic:
    # - if step gets a valid relation - lots of logic to extract correct structure
    # - else if step gets a list/tuple - recurse on each item
    # - else must be a single item - use realise_single_node
    #
    # - plus some helpful error handling
    #
    # - note: heavy lifting is passed to GraphDelta constructor - this function just adds IDs, nothing else
    

    def step(x, allow_scalar):
        if isinstance(x, ListOrTuple):
            if is_valid_relation_template(x):
                if len(x) == 2:
                    ent = step(x[0], True)
                    triples = [step((Z[ent], *tup), False) for tup in x[1]]
                    doubles = [x[1:] for x in triples]

                    return (ent, doubles)

                if len(x) == 3:
                    if isinstance(x[0], ListOrTuple):
                        assert is_valid_single_node(x[2])
                        item = step(x[2], True)
                        triples = [step((source, x[1], Z[item]), False) for source in x[0]]
                        sources = [x[0] for x in triples]
                        # Note: have to return None for the relation, as there are actually multiple relations
                        return (sources, None, item)

                    if isinstance(x[2], ListOrTuple):
                        assert is_valid_single_node(x[0])
                        item = step(x[0], True)
                        triples = [step((Z[item], x[1], target), False) for target in x[2]]
                        targets = [x[2] for x in triples]
                        # Note: have to return None for the relation, as there are actually multiple relations
                        return (item, None, targets)

                    # it's a relation triple
                    src_id = step(x[0], True)
                    if get_absorbed_id(x[1]) is not None:
                        raise Exception("Not allowing ids in a shorthand syntax transaction.")
                    rel_id = gen_id()
                    trg_id = step(x[2], True)
                    gd_exprs.append( (Z[src_id], x[1][rel_id], Z[trg_id]) )

                    return (src_id, rel_id, trg_id)
                raise Exception("Shouldn't get here")
            else:
                # recurse: return isomorphic structure with IDs
                return tuple((step(el, False) for el in x))
        
        # These next few ifs are for checks on syntax only
        if type(x) in shorthand_scalar_types:
            if not allow_scalar:
                raise Exception("Scalars are not allowed on their own to avoid accidental typos such as (ET.X, ET.Y, 'z') when (ET.X, RT.Y, 'z') is meant. If you want this behaviour, then create an explicit AET, i.e. (AET.String <= 'z').")
        
        iid,exprs = realise_single_node(x, gen_id)
        gd_exprs.extend(exprs)
        return iid

    step_res = step(xx, False)
    return step_res, construct_commands(gd_exprs, gen_id)


def unpack_receipt(unpacking_template, receipt: dict):
    def step(x):
        if isinstance(x, tuple):
            return tuple((step(el) for el in x))
        if isinstance(x, list):
            return [step(el) for el in x]
        if isinstance(x, dict):
            return {k: step(v) for k,v in x.items()}
        return receipt[x] if isinstance(x, str) or is_a(x, uid) or is_a(x, Delegate) else x
    return step(unpacking_template)



##################################################
# * Performing transaction
#------------------------------------------------

def perform_transaction_commands(commands: list, g: Graph):
    d_raes = {}  # keep track of instantiated RAEs with temp ids            
    try:
        with Transaction(g) as tx_now:
            # TODO: Have to change the behavior of Transaction(g) later I suspect
            frame_now = GraphSlice(tx_now)
            d_raes['tx'] = tx_now

            next_print = now()+5*seconds
            for i,cmd in enumerate(commands):
                if gd_perform_transaction_debug:
                    log.debug("Performing", i=i, cmd=cmd)
                if gd_timing:
                    if now() > next_print:
                        log.debug("Perform", i=i, total=len(commands))
                        next_print = now() + 5*seconds

                zz = None
                
                # print(f"{i}/{len(g_delta.commands)}: {g.graph_data.write_head * 16 / 1024 / 1024} MB")
                if cmd['cmd'] == 'instantiate' and type(cmd['rae_type']) in {EntityType, AttributeEntityType}:
                    zz = instantiate(cmd['rae_type'], g)
                
                elif cmd['cmd'] == 'instantiate' and type(cmd['rae_type']) in {RelationType}:
                    zz = instantiate(to_ezefref(d_raes[cmd['source']]), cmd['rae_type'], to_ezefref(d_raes[cmd['target']]), g) | in_frame[frame_now] | collect
                
                elif cmd['cmd'] == 'instantiate_value_node':
                    zz = instantiate_value_node_imp(cmd['value'], g)
                
                elif cmd['cmd'] == 'assign':
                    this_id = cmd['internal_id']
                    zz = d_raes.get(this_id, None)
                    if zz is None:
                        zz = most_recent_rae_on_graph(this_id, g)
                    assert zz is not None
                    if 'value' in cmd:
                        if zz | value | collect != cmd['value']:
                            # print("Assigning value of ", cmd['value'], "to a", AET(z))
                            internals.assign_value_imp(zz, cmd['value'])
                    else:
                        raise Exception("Assignment without an value")
                            
                    zz = now(zz)

                elif cmd['cmd'] == 'set_field':
                    z_source = d_raes.get(cmd['source_id'], None)
                    if z_source is None:
                        raise KeyError(f"set_field called with unknown source {cmd['source_id']}")

                    if 'value' in cmd:
                        # AET path
                        aet = map_scalar_to_aet_type[type(cmd['value'])](cmd['value'])
                    else:
                        # Entity path
                        z_target = d_raes.get(cmd['target_id'], None)
                        if z_target is None:
                            raise KeyError("set_field called with entity that is not known {cmd['target_id']}")

                    if cmd['incoming']:
                        opts = z_source | in_rels[cmd['rt']] | collect
                    else:
                        opts = z_source | out_rels[cmd['rt']] | collect
                    if len(opts) == 2:
                        raise Exception(f"Can't set_field to {z_source} because it has two or more relations of kind {rt}")
                    elif len(opts) == 1:
                        zz = single(opts)
                        if 'value' in cmd:
                            # AE path - overwrite value
                            if cmd['incoming']:
                                ae = source(zz)
                            else:
                                ae = target(zz)
                            if aet != rae_type(ae):
                                raise Exception(f"Can't fill or attach {ae} because it is the wrong type for value {cmd['value']}")
                            if value(ae) != cmd['value']:
                                internals.assign_value_imp(ae, cmd['value'])
                        else:
                            # Entity path - replace the relation
                            from ..pyzef import zefops as pyzefops
                            pyzefops.terminate(zz)
                            if cmd['incoming']:
                                zz = instantiate(z_target, cmd['rt'], z_source, g)
                            else:
                                zz = instantiate(z_source, cmd['rt'], z_target, g)
                    else:
                        if 'value' in cmd:
                            # AE path
                            ae = instantiate(aet, g)
                            internals.assign_value_imp(ae, cmd['value'])
                            if cmd['incoming']:
                                zz = instantiate(ae, cmd['rt'], z_source, g)
                            else:
                                zz = instantiate(z_source, cmd['rt'], ae, g)
                        else:
                            # Entity path
                            if cmd['incoming']:
                                zz = instantiate(z_target, cmd['rt'], z_source, g)
                            else:
                                zz = instantiate(z_source, cmd['rt'], z_target, g)
                    # This zz is the relation that connects the source/target
                    zz = now(zz)
                    
                elif cmd['cmd'] == 'merge':
                    # It is either an instance (with an 'origin_rae_uid' specified)
                    # or a delegate[...]
                    if is_a(cmd['origin_rae'], Delegate):
                        d = cmd['origin_rae']
                        zz = internals.delegate_to_ezr(d, g, True, 0)
                        zz = now(zz)
                    else:
                        candidate = most_recent_rae_on_graph(uid(cmd['origin_rae']), g)
                        if candidate is not None:
                            # this is already on the graph. Just assert and move on
                            assert abstract_type(cmd['origin_rae']) == abstract_type(candidate)
                            zz = candidate
                        else:
                            origin_rae_uid = uid(cmd['origin_rae'])
                            if isinstance(cmd['origin_rae'], Entity):
                                zz = internals.merge_entity_(
                                    g, 
                                    rae_type(cmd['origin_rae']), 
                                    origin_rae_uid.blob_uid,
                                    origin_rae_uid.graph_uid,
                                )
                            elif isinstance(cmd['origin_rae'], AttributeEntity):
                                zz = internals.merge_atomic_entity_(
                                    g, 
                                    rae_type(cmd['origin_rae']),
                                    origin_rae_uid.blob_uid,
                                    origin_rae_uid.graph_uid,
                                )
                            elif isinstance(cmd['origin_rae'], Relation):
                                src_origin_uid,_,trg_origin_uid = cmd['origin_rae'].d["uids"]
                                z_src = d_raes.get(src_origin_uid, most_recent_rae_on_graph(src_origin_uid, g))                                    
                                z_trg = d_raes.get(trg_origin_uid, most_recent_rae_on_graph(trg_origin_uid, g))                                    

                                assert z_src is not None
                                assert z_trg is not None                                    
                                zz = internals.merge_relation_(
                                    g, 
                                    rae_type(cmd['origin_rae']),
                                    to_ezefref(z_src),
                                    to_ezefref(z_trg),
                                    origin_rae_uid.blob_uid,
                                    origin_rae_uid.graph_uid,
                                )
                            # Special handling for TXNode and Root currently.
                            elif isinstance(cmd['origin_rae'], (TXNode, Root)):
                                raise Exception("Can only merge TXNode and Root onto the same graph")
                            else:
                                raise NotImplementedError

                        zz = now(zz)
                        if 'origin_rae' in cmd:
                            d_raes[uid(cmd['origin_rae'])] = zz
                
                elif cmd['cmd'] == 'terminate':
                    if uid(cmd['origin_rae']) not in g:
                        # raise KeyError(f"terminate {cmd['origin_rae']} called, but this RAE is not know to graph {uid(g)}")
                        # This is no longer an error, it should be ignored silently.
                        pass
                    else:
                        zz = most_recent_rae_on_graph(uid(cmd['origin_rae']), g)
                        if zz is not None:
                            from ..pyzef import zefops as pyzefops
                            pyzefops.terminate(zz)

                            zz = now(zz, allow_tombstone)
                elif cmd['cmd'] == 'tag':
                    if 'origin_rae' in cmd:
                        if uid(cmd['origin_rae']) not in g:
                            raise KeyError(f"tag {cmd['origin_rae']} called, but this RAE is not know to graph {g|uid}")
                        zz = most_recent_rae_on_graph(uid(cmd['origin_rae']), g)
                    else:
                        zz = d_raes[cmd['internal_id']]
                    # Go directly to the tagging instead of through another effect
                    pyzef.main.tag(zz, cmd['tag_name'], True)
                else:
                    raise RuntimeError(f'---------Unexpected case in performing graph delta tx: {cmd}')

                for this_id in get_ids(cmd):
                    d_raes[this_id] = zz

        # There is a weird edge case here - if a node is asked to be merged
        # in, but already existed on the graph AND there were no changes,
        # causing the transaction to be rolled-back, then the reference
        # frame of the ZefRefs will be wrong.

        # We check this by seeing if the index of the tx zefref is invalid.
        if index(d_raes['tx']) > g.graph_data.read_head:
            # Signal this by setting the tx to none (i.e. there was no tx)
            d_raes['tx'] = None

            # Update all ZefRefs to be in the latest frame instead of the previously created tx.
            reset_frame = in_frame[now(g)][allow_tombstone]
            for key,val in d_raes.items():
                if key == 'tx': continue

                # A terminated non-existent entity can return None
                if val is not None:
                    d_raes[key] = reset_frame(val)

    except Exception as exc:        
        raise RuntimeError(f"Error executing graph delta transaction exc={exc}") from exc
        
    return d_raes    # the transaction receipt

################################
# * General utils
#------------------------------

class TaggedVal:
    def __init__(self, arg, iid):
        if type(arg) in scalar_types:
            self.primitive = arg
        else:
            self.primitive = SerializedValue.serialize(arg)
        self.iid = iid


scalar_types = {int, float, bool, str, Time, QuantityFloat, QuantityInt, ZefEnumValue, SerializedValue, ValueType_}
shorthand_scalar_types = {int, float, bool, str, Time, QuantityFloat, QuantityInt, ZefEnumValue, SerializedValue}

def make_enum_aet(x):
    """ hacky work around function for now:
    e.g. given an enum value "EN.Color.White"
    we want to convert to the AET type expression: "AET.Enum.Color"
    """
    enum_typename: str = x.enum_type
    return getattr(AET.Enum, enum_typename)

def make_qf_aet(x):
    """ hacky work around function for now:
    e.g. given an enum value "QuantityFloat(2.1, EN.Unit.kilogram)"
    we want to convert to the AET type expression: "AET.QuantityFloat.kilogram"
    """
    quantity_unit: str = x.unit.enum_value
    return getattr(AET.QuantityFloat, quantity_unit)

def make_qi_aet(x):
    """ hacky work around function for now:
    e.g. given an enum value "QuantityInt(2.1, EN.Unit.kilogram)"
    we want to convert to the AET type expression: "AET.QuantityInt.kilogram"
    """
    quantity_unit: str = x.unit.enum_value
    return getattr(AET.QuantityInt, quantity_unit)

map_scalar_to_aet_type = {
    int:                lambda x: AET.Int,
    float:              lambda x: AET.Float,
    bool:               lambda x: AET.Bool,
    str:                lambda x: AET.String,
    Time:               lambda x: AET.Time,
    ZefEnumValue:       make_enum_aet,
    QuantityFloat:      make_qf_aet, 
    QuantityInt:        make_qi_aet, 
    ValueType_:         lambda x: AET.Type,
    }

def get_curried_arg(op, n):
    """ 
    utility function to get 
    e.g. get_curried_arg(Z['id1'][42], 0)       # => 'id1'
    e.g. get_curried_arg(Z['id1'][42], 1)       # => 42
    """
    # return op.el_ops[arg_indx][1][n]
    return LazyValue(op) | absorbed | nth[n] | collect



def most_recent_rae_on_graph(origin_uid: str, g: Graph)->ZefRef:
    """
    Will definitely not return a BT.ForeignInstance, always an instance.
    It could be that the node asked for has its origin on this graph 
    (the original rae may still be alive or it may be terminated)

    Args:
        origin_uid (str): the uid of the origin rae we are looking for
        g (Graph): on which graph are we looking?

    Returns:
        ZefRef: this graph knows about this: found instance
        None: this graph knows nothing about this RAE
    """
    if origin_uid not in g:
        return None     # this graph never knew about a RAE with this origin uid

    zz = g[origin_uid]
    if BT(zz) in {BT.FOREIGN_ENTITY_NODE, BT.FOREIGN_ATTRIBUTE_ENTITY_NODE, BT.FOREIGN_RELATION_EDGE}:
        from .graph_slice import get_instance_rae
        return get_instance_rae(origin_uid, now(g))
        
    elif BT(zz) in {BT.ENTITY_NODE, BT.ATTRIBUTE_ENTITY_NODE, BT.RELATION_EDGE}:
        if zz | exists_at[now(g)] | collect:
            return zz | in_frame[now(g)] | collect
        else:
            return None
    elif BT(zz) in {BT.ROOT_NODE, BT.TX_EVENT_NODE}:
        return zz | in_frame[now(g)] | collect
    else:
        raise RuntimeError("Unexpected option in most_recent_rae_on_graph")
        



# Generation of IDs and being able to detect whether an ID was auto-generated.

generated_prefix = "tmp_id_"
def make_generate_id():
    auto_id_counter = 0
    def generate_id():
        """count up"""
        nonlocal auto_id_counter
        auto_id_counter += 1
        return f"{generated_prefix}{auto_id_counter}"
    return generate_id

def is_generated_id(x):
    if not isinstance(x, str):
        return False
    return x.startswith(generated_prefix)

def filter_temporary_ids(receipt):
    return (receipt
            | items
            | filter[first | Not[is_generated_id]]
            | func[dict]
            | collect)

@func
def get_absorbed_id(obj):
    # THIS SHOULDN"T BE NEEDED! FIX!
    if is_a(obj, RT) or is_a(obj, ZefOp):
        obj = LazyValue(obj)
    return obj | absorbed | single_or[None] | collect


def equal_identity(a, b):
    if type(a) != type(b):
        return False

    # Things that don't produce the same object even though their python objects
    # are the same
    # if type(a) in [EntityType, RelationType, AttributeEntityType, ZefEnumValue]:
    #     return False

    return a == b

# @func
# def merge_no_overwrite(a,b):
#     d = {**a}

#     for k,v in b.items():
#         if k in d and not equal_identity(d[k], v):
#             raise Exception(f"The internal id '{k}' refers to multiple objects, including '{d[k]}' and '{v}'. This is ambiguous and not allowed.")
#         d[k] = v
#     return d

@func
def merge_no_overwrite(a,b):
    # This version is mutating because otherwise it's too slow
    for k,v in b.items():
        if k in a and not equal_identity(a[k], v):
            raise Exception(f"The internal id '{k}' refers to multiple objects, including '{a[k]!r}' and '{v!r}'. This is ambiguous and not allowed.")
        a[k] = v
    return a



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._core import RT
from .op_structs import  evaluating, LazyValue, Awaitable, ZefOp, CollectingOp, SubscribingOp, ForEachingOp

def register_zefop(rt, imp, tp):
    from .dispatch_dictionary import _op_to_functions
    import zef
    import re

    assert(type(rt) == zef.pyzef.main.RelationType)

    op_name = '_'.join(re.findall('[A-Z][^A-Z]*', str(rt))).lower()
    op = ZefOp(((rt, ()),))
    globals()[op_name] = op
    _op_to_functions[rt] = (imp, tp)
    return op


for_each        = ForEachingOp(ZefOp(()))
collect         = CollectingOp(ZefOp(()))
subscribe       = SubscribingOp(ZefOp(()))

run             = ZefOp(((RT.Run, (evaluating,)), ))

def make_zefop(rt):
    return ZefOp(((rt, ()), ))

cycle           = make_zefop(RT.Cycle)
repeat          = make_zefop(RT.Repeat)
contains        = make_zefop(RT.Contains)
contained_in    = make_zefop(RT.ContainedIn)
skip            = make_zefop(RT.Skip)
all             = make_zefop(RT.All)
any             = make_zefop(RT.Any)
slice           = make_zefop(RT.Slice)                 # lowercase slice is used in core Python
join            = make_zefop(RT.Join)
trim            = make_zefop(RT.Trim)
trim_left       = make_zefop(RT.TrimLeft)
trim_right      = make_zefop(RT.TrimRight)
yo              = make_zefop(RT.Yo)
tap             = make_zefop(RT.Tap)
push            = make_zefop(RT.Push)

cartesian_product = make_zefop(RT.CartesianProduct)
permutations    = make_zefop(RT.Permutations)
combinations    = make_zefop(RT.Combinations)
sum             = make_zefop(RT.Sum)
product         = make_zefop(RT.Product)
add             = make_zefop(RT.Add)
subtract        = make_zefop(RT.Subtract)
multiply        = make_zefop(RT.Multiply)
divide          = make_zefop(RT.Divide)
mean            = make_zefop(RT.Mean)
variance        = make_zefop(RT.Variance)
power           = make_zefop(RT.Power)
exponential     = make_zefop(RT.Exponential)
logarithm       = make_zefop(RT.Logarithm)

equals          = make_zefop(RT.Equals)
greater_than    = make_zefop(RT.GreaterThan)
less_than       = make_zefop(RT.LessThan)
greater_than_or_equal = make_zefop(RT.GreaterThanOrEqual)
less_than_or_equal = make_zefop(RT.LessThanOrEqual)
Not             = make_zefop(RT.Not)
And             = make_zefop(RT.And)
Or              = make_zefop(RT.Or)
xor             = make_zefop(RT.Xor)

always          = make_zefop(RT.Always)
docstring       = make_zefop(RT.Docstring)
source_code     = make_zefop(RT.SourceCode)

absorbed        = make_zefop(RT.Absorbed)
without_absorbed= make_zefop(RT.WithoutAbsorbed)

get_in          = make_zefop(RT.GetIn)
insert_in       = make_zefop(RT.InsertIn)
remove_in       = make_zefop(RT.RemoveIn)
update_in       = make_zefop(RT.UpdateIn)
update_at       = make_zefop(RT.UpdateAt)
insert_at       = make_zefop(RT.InsertAt)
update          = make_zefop(RT.Update)
remove_at       = make_zefop(RT.RemoveAt)
merge           = make_zefop(RT.Merge)
merge_with      = make_zefop(RT.MergeWith)
int_to_alpha    = make_zefop(RT.IntToAlpha)
permute_to      = make_zefop(RT.PermuteTo)


# Implemented Lazy ZefOps
expect          = make_zefop(RT.Expect)
filter          = make_zefop(RT.Filter)
without         = make_zefop(RT.Without)
select_keys     = make_zefop(RT.SelectKeys)
modulo          = make_zefop(RT.Modulo)
select_by_field = make_zefop(RT.SelectByField)
apply_functions = make_zefop(RT.ApplyFunctions)
map             = make_zefop(RT.Map)
map_cat         = make_zefop(RT.MapCat)
identity        = make_zefop(RT.Identity)
concat          = make_zefop(RT.Concat)
zip             = make_zefop(RT.Zip)
prepend         = make_zefop(RT.Prepend)
append          = make_zefop(RT.Append)
interleave      = make_zefop(RT.Interleave)
interleave_longest = make_zefop(RT.InterleaveLongest)
chunk           = make_zefop(RT.Chunk)
sliding         = make_zefop(RT.Sliding)
stride          = make_zefop(RT.Stride)
insert          = make_zefop(RT.Insert)
insert_into     = make_zefop(RT.InsertInto)
reverse_args    = make_zefop(RT.ReverseArgs)
remove          = make_zefop(RT.Remove)
get             = make_zefop(RT.Get)
get_field       = make_zefop(RT.GetField)
enumerate       = make_zefop(RT.Enumerate)
items           = make_zefop(RT.Items)
values          = make_zefop(RT.Values)
keys            = make_zefop(RT.Keys)
reverse         = make_zefop(RT.Reverse)
reduce          = make_zefop(RT.Reduce)
iterate         = make_zefop(RT.Iterate)
scan            = make_zefop(RT.Scan)
group_by        = make_zefop(RT.GroupBy)
group           = make_zefop(RT.Group)
transpose       = make_zefop(RT.Transpose)
frequencies     = make_zefop(RT.Frequencies)
max             = make_zefop(RT.Max)
min             = make_zefop(RT.Min)
max_by          = make_zefop(RT.MaxBy)
min_by          = make_zefop(RT.MinBy)
clamp           = make_zefop(RT.Clamp)
first           = make_zefop(RT.First)
second          = make_zefop(RT.Second)
last            = make_zefop(RT.Last)
single          = make_zefop(RT.Single)           
single_or       = make_zefop(RT.SingleOr)
only            = make_zefop(RT.Single)                 # TODO: retire, since we renamed this to 'single'
identity        = make_zefop(RT.Identity)
take            = make_zefop(RT.Take)
take_while      = make_zefop(RT.TakeWhile)
take_while_pair = make_zefop(RT.TakeWhilePair)
take_until      = make_zefop(RT.TakeUntil)                 # TODO: use 'take_until' in the RX-sense: complete the stream based on another stream emitting an item. Add 'including' to take_while as a flag for the current behavior?
skip_while      = make_zefop(RT.SkipWhile)
skip_until      = make_zefop(RT.SkipUntil)
skip            = make_zefop(RT.Skip)
length          = make_zefop(RT.Length) 
nth             = make_zefop(RT.Nth) 
now             = make_zefop(RT.Now) 
events          = make_zefop(RT.Events) 
preceding_events= make_zefop(RT.PrecedingEvents) 
to_delegate     = make_zefop(RT.ToDelegate) 
delegate_of     = make_zefop(RT.DelegateOf) 
target          = make_zefop(RT.Target) 
source          = make_zefop(RT.Source) 
L               = make_zefop(RT.L)
O               = make_zefop(RT.O)
Z               = make_zefop(RT.Z)
RAE             = make_zefop(RT.RAE)
time            = make_zefop(RT.Time) 
value           = make_zefop(RT.Value)
sort            = make_zefop(RT.Sort)
uid             = make_zefop(RT.Uid)
frame           = make_zefop(RT.Frame)
discard_frame   = make_zefop(RT.DiscardFrame)
to_frame        = make_zefop(RT.InFrame)                           # TODO: retire this. Use 'in_frame' instead
in_frame        = make_zefop(RT.InFrame)
to_graph_slice  = make_zefop(RT.ToGraphSlice)
to_tx           = make_zefop(RT.ToTx)
time_travel     = make_zefop(RT.TimeTravel)
next_tx         = make_zefop(RT.NextTX)             
previous_tx     = make_zefop(RT.PreviousTX)     
to_ezefref      = make_zefop(RT.ToEZefRef)
root            = make_zefop(RT.Root)
terminate       = make_zefop(RT.Terminate) 
assign          = make_zefop(RT.Assign) 
is_a            = make_zefop(RT.IsA)
is_represented_as = make_zefop(RT.IsRepresentedAs)
representation_type = make_zefop(RT.RepresentationType)
rae_type        = make_zefop(RT.RaeType)
abstract_type   = make_zefop(RT.AbstractType)
fill_or_attach  = make_zefop(RT.FillOrAttach)
set_field       = make_zefop(RT.SetField)
Assert          = make_zefop(RT.Assert)
allow_tombstone = make_zefop(RT.AllowTombstone)
tag             = make_zefop(RT.Tag)
untag           = make_zefop(RT.Untag)
sync            = make_zefop(RT.Sync)
to_clipboard    = make_zefop(RT.ToClipboard)
from_clipboard  = make_zefop(RT.FromClipboard)
text_art        = make_zefop(RT.TextArt)

sign            = make_zefop(RT.Sign)
attempt         = make_zefop(RT.Attempt)
bypass          = make_zefop(RT.Bypass)
pattern         = make_zefop(RT.Pattern)
replace         = make_zefop(RT.Replace)
distinct        = make_zefop(RT.Distinct)
distinct_by     = make_zefop(RT.DistinctBy)
is_distinct     = make_zefop(RT.IsDistinct)
is_distinct_by  = make_zefop(RT.IsDistinctBy)
shuffle         = make_zefop(RT.Shuffle)
split           = make_zefop(RT.Split)
split_left      = make_zefop(RT.SplitLeft)
split_right     = make_zefop(RT.SplitRight)
graphviz        = make_zefop(RT.Graphviz)

blueprint       = make_zefop(RT.Blueprint)
exists_at       = make_zefop(RT.ExistsAt)
aware_of        = make_zefop(RT.AwareOf)
base_uid        = make_zefop(RT.BaseUid)
origin_uid      = make_zefop(RT.OriginUid)
origin_rae      = make_zefop(RT.OriginRAE)

has_out         = make_zefop(RT.HasOut)                # z1 | has_out[RT.Foo] use  (z1, RT.Foo, Z) | exists  /   (z, RT.Foo, RAE) | exists[g]  /   (z, RT.Foo, RAE) | exists[now(g)][single]
has_in          = make_zefop(RT.HasIn)                 # z1 | has_in[RT.Foo]  use  (Z, RT.Foo, z1) | exists

In              = make_zefop(RT.In)
Ins             = make_zefop(RT.Ins)
Out             = make_zefop(RT.Out)
Outs            = make_zefop(RT.Outs)
ins_and_outs    = make_zefop(RT.InsAndOuts)
in_rel          = make_zefop(RT.InRel)
in_rels         = make_zefop(RT.InRels)
out_rel         = make_zefop(RT.OutRel)
out_rels        = make_zefop(RT.OutRels)


is_zefref_promotable= make_zefop(RT.IsZefRefPromotable)  # Retire this. this is a old love level operator. We can use is_a[RAE] or an extended concept new.
time_slice      = make_zefop(RT.TimeSlice)        
graph_slice_index=make_zefop(RT.GraphSliceIndex) 
    
instantiation_tx= make_zefop(RT.InstantiationTx)       # use tx[instantiated]
termination_tx  = make_zefop(RT.TerminationTx)         # use tx[terminated]   
relations       = make_zefop(RT.Relations)             # g | now | all[(z1, RT.Bar, z2)]   with pattern matching style any of the three args can also be replaced with a more general class
relation        = make_zefop(RT.Relation)              # looking through our code base for use cases of this op, I don't think a separate operator is necessary. Just use the syntax above followed by ... | single. If required more often, it is much easier to add this in future than to remove it.
unpack          = make_zefop(RT.Unpack)
_any            = make_zefop(RT._Any)                  # used as a wildcard
has_relation    = make_zefop(RT.HasRelation)     

replace_at      = make_zefop(RT.ReplaceAt)           
pad_left        = make_zefop(RT.PadLeft)           
pad_right       = make_zefop(RT.PadRight)           
pad_center      = make_zefop(RT.PadCenter)           
ceil            = make_zefop(RT.Ceil)           
floor           = make_zefop(RT.Floor)           
round           = make_zefop(RT.Round)           
random_pick     = make_zefop(RT.RandomPick)           

to_json         = make_zefop(RT.ToJSON)           
from_json       = make_zefop(RT.FromJSON)   

to_yaml         = make_zefop(RT.ToYaml)
from_yaml       = make_zefop(RT.FromYaml)

to_toml         = make_zefop(RT.ToToml)
from_toml       = make_zefop(RT.FromToml)

to_csv         = make_zefop(RT.ToCSV)
from_csv       = make_zefop(RT.FromCSV)

read_file      = make_zefop(RT.ReadFile)
load_file      = make_zefop(RT.LoadFile)
write_file     = make_zefop(RT.WriteFile)
save_file      = make_zefop(RT.SaveFile)


pandas_to_gd = make_zefop(RT.PandasToGd)

to_pipeline    = make_zefop(RT.ToPipeline)
inject         = make_zefop(RT.Inject)
inject_list    = make_zefop(RT.InjectList)


is_alpha        = make_zefop(RT.IsAlpha)
is_numeric      = make_zefop(RT.IsNumeric)
is_alpha_numeric= make_zefop(RT.IsAlphaNumeric)
to_upper_case   = make_zefop(RT.ToUpperCase)
to_lower_case   = make_zefop(RT.ToLowerCase)

to_pascal_case  = make_zefop(RT.ToPascalCase)
to_camel_case   = make_zefop(RT.ToCamelCase)
to_kebab_case   = make_zefop(RT.ToKebabCase)
to_snake_case   = make_zefop(RT.ToSnakeCase)
to_screaming_snake_case = make_zefop(RT.ToScreamingSnakeCase)


make_request  = make_zefop(RT.MakeRequest)


zascii_to_asg       = make_zefop(RT.ZasciiToAsg)
zascii_to_flatgraph = make_zefop(RT.ZasciiToFlatGraph)
zascii_to_blueprint_fg = make_zefop(RT.ZasciiToBlueprintFg)

                # Syntax????????????????? 
                # has_relation(z1, RT.Foo, z2)      replaced by 
                #       1) (z1, RT.Foo, z2) | exists[g_slice]   or 
                #       1) (z1, RT.Foo, z2) | contained_in[g_slice]   or 
                #       2) g_slice | contains[(z1, RT.Foo, z2)]
# Syntax choices:   
#       exists or contained_in?
#       All or instances?    Also: my_delegate | all    or my_delegate | instances?





merge           = make_zefop(RT.Merge)                 # We need this for observables. Only there?

blake3          = make_zefop(RT.Blake3)
value_hash      = make_zefop(RT.ValueHash)        
to_zef_list     = make_zefop(RT.ToZefList)
transact        = make_zefop(RT.Transact)
# transact

# subscribe
# keep_alive
# incoming
# outgoing
# on_instantiation
# on_termination
# on_value_assignment


peel            = make_zefop(RT.Peel)                
match           = make_zefop(RT.Match)                
Range           = make_zefop(RT.Range)      
zstandard_compress = make_zefop(RT.ZstandardCompress)
zstandard_decompress = make_zefop(RT.ZstandardDecompress)
to_bytes        = make_zefop(RT.ToBytes)
utf8bytes_to_string  = make_zefop(RT.Utf8bytesToString)
base64string_to_bytes = make_zefop(RT.Base64stringToBytes)
bytes_to_base64string = make_zefop(RT.BytesToBase64string)
is_between      = make_zefop(RT.IsBetween)
If              = make_zefop(RT.If)

field           = make_zefop(RT.Field)
fields          = make_zefop(RT.Fields)
apply           = make_zefop(RT.Apply)
split_on_next   = make_zefop(RT.SplitOnNext)
indexes_of      = make_zefop(RT.IndexesOf)
to_flatgraph    = make_zefop(RT.ToFlatGraph)
parse           = make_zefop(RT.Parse)


examples        = make_zefop(RT.Examples)
signature       = make_zefop(RT.Signature)
tags            = make_zefop(RT.Tags)
related_ops     = make_zefop(RT.RelatedOps)
operates_on     = make_zefop(RT.OperatesOn)
used_for        = make_zefop(RT.UsedFor)


# TODO: implement
on              = make_zefop(RT.On)         
gather          = make_zefop(RT.Gather)
alias           = make_zefop(RT.Alias)
splice          = make_zefop(RT.Splice)
flatten_dict    = make_zefop(RT.FlattenDict)
unflatten_dict  = make_zefop(RT.UnflattenDict)

          
# match
# split_before
# split_after
# split_at
# split_when



# delay
# window(Max[10], Max[30/sec, over[2*sec]])
# 
# time_travel         # RuntimeError: Only(EZefRefs zs) request, but length was 0





# -------- These are not ZefOps, but using the `.` operator, they return ZefOps.
# The user may not even be aware of this distinction and therefore this namespace 
# is the most natural to put them in.
class FClass:    
    def __getattr__(self, s: str):        
        return field[RT(s)]    # just returns a normal zefop called 'field'

F = FClass()

class FsClass:    
    def __getattr__(self, s: str):        
        return fields[RT(s)]    # just returns a normal zefop called 'field'

Fs = FsClass()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...pyzef.internals import login, login_manual, logout
from ..error import Error
from .fx_types import Effect

def zefhub_login_handler(eff: Effect):
    try:
        jupyter_shell = False
        try:
            import IPython
            shell = IPython.get_ipython()
            if shell is not None and shell.__class__.__name__ != 'TerminalInteractiveShell':
                jupyter_shell = True
        except:
            pass
        comments = """=================================
    You are now logged in to ZefHub. You can synchronize graphs which will enable them to be stored on
    ZefHub. Any ETs, RT, ENs and KWs you create will also be synchronized with ZefHub.

    Note: your credentials have been stored in the zef config directory.
    By default these will be used to automatically connect to ZefHub on import of the zef module.
    If you would like to change this behavior, please see the `config` zefop for more information.
    ================================="""

        if 'auth_key' in eff:
            login_manual(eff['auth_key'])
            comments = """=================================
    You are logged in as a guest user, which allows you to view public graphs but
    does not allow for synchronising new graphs with ZefHub.

    Disclaimer: any ETs, RTs, ENs, and KWs that you query will be stored with ZefHub.
    ================================="""
        else:
            if jupyter_shell:
                print("=================================")
                print("This command starts a local libzef login web server. If you are running inside")
                print("of a jupyter notebook then this will start the auth prompt on the server where")
                print("jupyter is running.")
                print()
                print("If you are using jupyter through a Google Colab server or a different JupyterHub")
                print("server, then this will not be useful to you. In fact, we do not recommend you")
                print("login with your credentials on a remote server.")
                print()
                print("You can instead login as a guest user using `login_as_guest | run`")
                print()
                print("Advanced users: If you are sure you would like to login with your credentials,")
                print("then look for workarounds involving environment variables and manual login commands.")
                print("=================================")
            login()

        if jupyter_shell:
            print(comments)
            # TODO: Report user account here
            return {"success": True}
        else:
            # TODO: Report user account here
            return {"success": True,
                    "comments": comments}
    except Exception as e:
        return Error(f'executing FX.ZefHub.Login for effect {eff}:\n{repr(e)}')


def zefhub_logout_handler(eff: Effect):
    try:
        logout()
        print("Warning, after logging out all graphs will have been unloaded. Any accesses of a ZefRef")
        print("referencing these old graphs will cause a segmentation fault!")
        return {"success": True}
    except Exception as e:
        return Error(f'executing FX.ZefHub.Login for effect {eff}:\n{repr(e)}')



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# These functions are here for __main__.py and whoever wants to write their own
# main equivalent.


from ... import *
from ...ops import *

from zef.core.logger import log

from .schema_file_parser import parse_partial_graphql, json_to_minimal_nodes

def create_schema_graph(schema_string, hooks_string=None):
    """Create a Zef representation of the GraphQL schema given in `schema_string`
along with the zef functions given in hooks_string on a new blank graph `g`.

    The graph `g` is marked keep alive. If the server is manually restarted in
    the same python session multiple times, then this should be removed by the
    calling function.
    """

    g_schema = Graph()
    set_keep_alive(g_schema, True)

    if hooks_string is not None:
        # Prep the schema graph with the hooks, so that the schema nodes can point directly at these
        prepare_hooks(g_schema, hooks_string)

    parsed_dict = parse_partial_graphql(schema_string)
    commands = json_to_minimal_nodes(parsed_dict, g_schema)
    r = commands | transact[g_schema] | run
    root = r["root"]

    return root


def prepare_hooks(graph, hooks_string):
    with Transaction(graph):
        # Note that we save the string to a temporary file, because @func requires
        # getsource which cannot identify the source from inside an arbitrary
        # string.
        import tempfile, os
        fd,path = tempfile.mkstemp(prefix="SimpleGQL_hooks_", suffix=".py")
        try:
            with open(fd, "w") as file:
                file.write(hooks_string)
            code = compile(hooks_string, path, "exec")
            globs = {"g": graph}
            exec(code, globs, globs)
        finally:
            os.unlink(path)

        # We could try and be fancy here with autodetection of names - but for the initial test let's just make the user do everything, including tagging nodes.

        # Going to try autodetecting any zef functions on the graph
        funcs = graph | now | all[ET.ZEF_Function] | collect

        nodup_and_dup_funcs = (funcs
                    | group_by[F.OriginalName]
                    | group_by[second
                                | length
                                | greater_than_or_equal[2]]
                            [[False,True]]
                    | func[dict]
                    | collect)

        nodups = nodup_and_dup_funcs[False] | map[apply[first, second | single]] | collect
        dups = nodup_and_dup_funcs[True] | map[first] | collect

        if len(dups) > 0:
            log.error("There are zef functions that have duplicate names on the schema graph. I can't autotag these functions and will ignore them.", dups=dups)

        for name,z in nodups:
            z | tag[name] | graph | run


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def c_schema_validator(ctx):
    """Don't call this explicitly. Only for the zefdb core."""

    # We can move this to another section later on
    from .. import Graph
    from .. import _ops as zo
    from ...pyzef.internals import AbortTransaction, RT

    gs = ctx | zo.to_graph_slice | zo.collect
    for schema in gs | zo.root | zo.Outs[RT.ZEF_Schema]:
        lt = schema | zo.value | zo.collect
        if {"graph_slice": gs} | zo.Not[zo.is_a[lt]] | zo.collect:
            raise Exception(f"{schema} failed to validate graph slice.")

def register_schema_validator():
    from ...pyzef import internals
    internals.register_schema_validator(c_schema_validator)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._core import *

# "Custom Entities" are also used to represent plain data values.
# Originally, we used ZefOps for these, which absorb other Zef Values
# in the same way as Entities do.
# Using ZefOps when just wanting plain data is not the best choice though.
# It does not convey the intent of what it is (a zefop acts like a function
# and is chainable). But most importantly, they are not recognized as values
# to act as the initial value that automatically act as the input value 
# when written at the beginning of a data pipeline, which requires constant 
# additional mental overhead.
# It may seem weird at first, but using Entities to represent pure data 
# expressions works fairly well. All the syntactic behavior and absorption 
# of elements already comes out of the box.
# Synchronization across multiple processes and over ZefHub is also already
# implemented. The ability to merge an entity into any graph and the lineage
# being taken care of is useful and one can also immediately treat 
# custom entities as first class citizens on a graph, that relations for 
# meta-information etc. can be added to.

# The one thing that is missing is that the repr output for data expressions
# should agree with the written expression for legibility. 
# For this reason a bit of logic is added to the repr function for Entities.
# If it is a custom entity (as indicated by a special graph uid), a special
# name to display may be presen in the _custom_entity_display_names dictionary.
# TODO: make _custom_entity_display_names part of a unified zef process state.

_custom_entity_display_names = {}

def get_custom_entity_name_dict():
    return _custom_entity_display_names


#                                     _     _           _                       _       ____      _     _____      ____  _                                                            
#                                    / \   | |__   ___ | |_  _ __   __ _   ___ | |_    |  _ \    / \   | ____|    / ___|| |  __ _  ___  ___   ___  ___                                
#   _____  _____  _____  _____      / _ \  | '_ \ / __|| __|| '__| / _` | / __|| __|   | |_) |  / _ \  |  _|     | |    | | / _` |/ __|/ __| / _ \/ __|    _____  _____  _____  _____ 
#  |_____||_____||_____||_____|    / ___ \ | |_) |\__ \| |_ | |   | (_| || (__ | |_    |  _ <  / ___ \ | |___    | |___ | || (_| |\__ \\__ \|  __/\__ \   |_____||_____||_____||_____|
#                                 /_/   \_\|_.__/ |___/ \__||_|    \__,_| \___| \__|   |_| \_\/_/   \_\|_____|    \____||_| \__,_||___/|___/ \___||___/                               
#                                                                                                                                                                                     


class Entity:
    """ 
    A value representation of an "abstract entity" (in the 
    sense of an abstract vs concrete object 
    https://plato.stanford.edu/entries/abstract-objects/).

    This object/value can be used to refer to an entity
    without the context of a reference frame (graph slice: 
    subject and subject's time) or a timeless observational
    frame (graph). An abstract entity itself is timeless.

    It is simply a wrapper around the entity's type and its
    origin's id (serving as an identifier for its lineage group.)

    This will become a zef value in future.
    """
    def __init__(self, x):
        from ._ops import origin_uid
        if isinstance(x, ZefRef) or isinstance(x, EZefRef):
            assert BT(x)==BT.ENTITY_NODE
            self.d = {
                'type': ET(x),
                'uid': origin_uid(x),
                'absorbed': (),
            }
        
        elif isinstance(x, Entity):
            return x

        elif isinstance(x, dict):
            assert 'type' in x and 'uid' in x
            assert type(x['type']) == EntityType
            x['absorbed'] = x.get('absorbed', ())
            self.d = x
        else:
            raise TypeError(f"can't construct an abstract entity from a type(x)={type(x)}.  Value passed in: x={x}")

    def __repr__(self):
            # whether something is a custom entity with a overloaded repr
            # is encoded in the graph uid
            custom_entity_display_names = get_custom_entity_name_dict()
            base_name: str = (
                custom_entity_display_names[self.d['uid'].blob_uid]
                if str(self.d['uid'])[16:] == '0000000000000001'
                else f'Entity({repr(self.d["type"])}, {repr(self.d["uid"])})' 
            )
            # also show any absorbed elements
            return base_name + ''.join(('[' + repr(el) + ']' for el in self.d['absorbed']))

    def __eq__(self, other):
        if not isinstance(other, Entity): return False
        return self.d['type'] == other.d['type'] and self.d['uid'] == other.d['uid'] and self.d['absorbed'] == other.d['absorbed']

    def __hash__(self):
        return hash(self.d['uid'])

    def __getitem__(self, x):
        # append x to absorbed (a tuple). Return a new object
        return Entity({**self.d, 'absorbed': (*self.d['absorbed'], x)})
        


    

class AttributeEntity:
    """ 
    A value representation of an "abstract atomic entity".
    This will become a zef value in future.
    """
    def __init__(self, x):
        from ._ops import origin_uid
        if isinstance(x, ZefRef) or isinstance(x, EZefRef):
            assert BT(x)==BT.ATTRIBUTE_ENTITY_NODE
            self.d = {
                'type': AET(x),
                'uid': origin_uid(x),
                'absorbed': (),
            }        
        elif isinstance(x, AttributeEntity):
            return x
        elif isinstance(x, dict):
            assert 'type' in x and 'uid' in x
            assert type(x['type']) == AttributeEntityType
            x['absorbed'] = x.get('absorbed', ())
            self.d = x
        else:
            raise TypeError(f"can't construct an abstract atomic entity from a type(x)={type(x)}.  Value passed in: x={x}")

    def __repr__(self):
        return f'AttributeEntity({repr(self.d["type"])}, {repr(self.d["uid"])})' + ''.join(('[' + repr(el) + ']' for el in self.d['absorbed']))

    def __eq__(self, other):
        if not isinstance(other, AttributeEntity): return False
        return self.d['type'] == other.d['type'] and self.d['uid'] == other.d['uid']

    def __le__(self, value):
        from ._ops import assign
        return self | assign[value]
    
    def __hash__(self):
        return hash(self.d['uid'])

    def __getitem__(self, x):
        return AttributeEntity({**self.d, 'absorbed': (*self.d['absorbed'], x)})
    

class Relation:
    """ 
    A value representation of an "abstract relation".
    This is a thin wrapper around the type (triple)

    It will become a zef value in future.
    """
    def __init__(self, x):
        from ._ops import origin_uid, rae_type, source, target
        if isinstance(x, ZefRef) or isinstance(x, EZefRef):
            assert BT(x)==BT.RELATION_EDGE
            self.d = {
                'type': (rae_type(source(x)), RT(x), rae_type(target(x))),
                'uids': (origin_uid(source(x)), origin_uid(x), origin_uid(target(x))),
                'absorbed': (),
            }        
        elif isinstance(x, Relation):
            return x
        elif isinstance(x, dict):
            assert 'type' in x and 'uids' in x
            assert type(x['type']) == tuple and len(x['type']) == 3 and type(x['type'][1]) == RelationType
            assert type(x['uids']) == tuple and len(x['uids']) == 3 
            x['absorbed'] = x.get('absorbed', ())
            self.d = x
        else:
            raise TypeError(f"can't construct an abstract relation from a type(x)={type(x)}.  Value passed in: x={x}")

    def __repr__(self):
        return f'Relation({repr(self.d["type"])}, {repr(self.d["uids"])})' + ''.join(('[' + repr(el) + ']' for el in self.d['absorbed']))

    def __eq__(self, other):
        if not isinstance(other, Relation): return False
        return self.d['type'] == other.d['type'] and self.d['uids'] == other.d['uids']

    def __hash__(self):
            return hash(''.join([str(x) for x in self.d['uids']]))
            
    def __getitem__(self, x):
        return Relation({**self.d, 'absorbed': (*self.d['absorbed'], x)})

class TXNode:
    """ 
    A value representation of an "abstract transaction".
    """
    def __init__(self, x):
        from ._ops import origin_uid
        if isinstance(x, ZefRef) or isinstance(x, EZefRef):
            assert BT(x)==BT.TX_EVENT_NODE
            self.d = {
                'uid': origin_uid(x),
                'absorbed': (),
            }
        
        elif isinstance(x, TXNode):
            self.d = {
                'uid': x.d['uid'],
                'absorbed': x.d['absorbed']
            }

        else:
            raise TypeError(f"can't construct an abstract transaction from a type(x)={type(x)}.  Value passed in: x={x}")

    def __repr__(self):
        # also show any absorbed elements
        base_name = f'TXNode({repr(self.d["uid"])})' 
        return base_name + ''.join(('[' + repr(el) + ']' for el in self.d['absorbed']))

    def __eq__(self, other):
        if not isinstance(other, TXNode): return False
        return self.d['uid'] == other.d['uid'] and self.d['absorbed'] == other.d['absorbed']

    def __hash__(self):
        return hash((TXNode, self.d['uid']))

    def __getitem__(self, x):
        # append x to absorbed (a tuple). Return a new object
        temp = TXNode(self)
        temp.d['absorbed'] = (*self.d['absorbed'], x)
        return temp
        

class Root:
    """ 
    A value representation of an "abstract root node".
    """
    def __init__(self, x):
        from ._ops import origin_uid
        if isinstance(x, ZefRef) or isinstance(x, EZefRef):
            assert BT(x)==BT.ROOT_NODE
            self.d = {
                'uid': origin_uid(x),
                'absorbed': (),
            }
        
        elif isinstance(x, Root):
            self.d = {
                'uid': x.d['uid'],
                'absorbed': x.d['absorbed']
            }

        else:
            raise TypeError(f"can't construct an abstract root node from a type(x)={type(x)}.  Value passed in: x={x}")

    def __repr__(self):
        # also show any absorbed elements
        base_name = f'Root({repr(self.d["uid"])})' 
        return base_name + ''.join(('[' + repr(el) + ']' for el in self.d['absorbed']))

    def __eq__(self, other):
        if not isinstance(other, Root): return False
        return self.d['uid'] == other.d['uid'] and self.d['absorbed'] == other.d['absorbed']

    def __hash__(self):
        return hash((Root, self.d['uid']))

    def __getitem__(self, x):
        # append x to absorbed (a tuple). Return a new object
        temp = Root(self)
        temp.d['absorbed'] = (*self.d['absorbed'], x)
        return temp


def abstract_rae_from_rae_type_and_uid(rae_type, uid):
    from ._ops import is_a
    if is_a(rae_type, ET):
        return Entity({"type": rae_type, "uid": uid})
    elif is_a(rae_type, AET):
        return AttributeEntity({"type": rae_type, "uid": uid})
    else:
        assert is_a(rae_type, RT)
        raise Exception("Unable to create an abstract Relation without knowing its source and target")
        







def make_custom_entity(name_to_display: str, predetermined_uid=None):
    import random
    from . import internals
    d = ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f')

    if predetermined_uid is not None:
        if not isinstance(predetermined_uid, str):
            raise TypeError('make_custom_entity must be called with a string')
        if len(predetermined_uid) != 16:
            raise ValueError('predetermined_uid must be of length 16')
        if not all((el in d for el in predetermined_uid)):
            raise ValueError('predetermined_uid may only contain hexadecimal digits (lower case)')
    
    def generate_uid():
        return ''.join([random.choice(d) for _ in range(16)])

    src_g_uid = internals.BaseUID('0000000000000001')   # special code for custom entities
    this_uid = internals.BaseUID(generate_uid() if predetermined_uid is None else predetermined_uid)

    if name_to_display is not None:
        custom_entity_display_names: dict = get_custom_entity_name_dict()
        if this_uid in custom_entity_display_names:
            raise KeyError(f"Error in make_custom_entity: the key {this_uid} was already registered in custom_entity_display_names.")
        custom_entity_display_names[this_uid] = name_to_display

    return Entity({
        'type': ET.ZEF_CustomEntity,
        'uid': internals.EternalUID(this_uid, src_g_uid),
    })







# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fx_types import Effect
from ..error import Error
import yaml 
import json 
import toml
import  pandas as pd
import io
from ..image import Image

#TODO Docstring!

def read_localfile_handler(eff: Effect):
    """
    Reads the file as a string. Returns a str object.
    To read as a binary file, use FX.LocalFile.ReadBinary.
    >>> FX.LocalFile.Read(filename='my_file.txt')

    Response example:
    {
        'content': some_dict,
        'filename': 'my_file.txt',
    }
    """
    try:
        filename  = eff["filename"]
        with open(filename, "r") as f:
            f = f.read()
        return {"content": f, "filename": filename}
    except Exception as e:
        raise RuntimeError(f"Error reading file in FX.LocalFile.Read for effect={eff}:\n {repr(e)}")

def readbinary_localfile_handler(eff: Effect):
    """
    Reads a localfile in binary mode and returns the content as bytes.
    >>> FX.LocalFile.ReadBinary(filename='my_file.txt')

    Response example:
    {
        'content': some_dict,
        'filename': 'my_file.txt',
    }
    """
    try:
        filename  = eff["filename"]
        with open(filename, "rb") as f:
            f = f.read()
        return {"content": f, "filename": filename}
    except Exception as e:
        raise RuntimeError(f"Error reading file in FX.LocalFile.Read for effect={eff}:\n {repr(e)}")



def load_localfile_handler(eff: Effect):
    """
    Opinionated loading of file: it parses based on the file extension.
    If you want plain reading of a file, use 
    >>> FX.LocalFile.Load(filename='my_file.json')

    Supported formats:
    - json
    - csv
    - toml
    - yaml / yml
    - svg
    - png
    - jpg
    - jpeg
    - gif

    Response example:
    {
        'content': some_dict,
        'format': 'json',
        'filename': 'my_file.json',
    }
    """
    try:
        filename  = eff["filename"]
        settings  = eff.get("settings", {})
        format    = eff.get("format", None)

        if not format:
            if "." not in filename: return Error.ValueError("filename is missing an extension and a format wasn't provided!", filename)
            else: format = filename[filename.rindex(".") + 1:]
        elif "." not in filename: filename = filename + f".{format}"


        if format in {"svg", "png", "jpg", "jpeg", "gif"}:
            with open(filename, "rb") as f:
                content = f.read()
            content = Image(content, format)
        else:
            with open(filename, "rb") as f:
                content = f.read()  
            if format in {"yaml", "yml"}:
                content = yaml.safe_load(content)
            elif format == "toml":
                content = toml.loads(content)
            elif format == "csv":
                content = pd.read_csv(io.StringIO(content), **settings)
            elif format == "json":
                content = json.loads(content)

        return {"content": content, "format": format, "filename": filename}
    except Exception as e:
        return Error(f'executing FX.LocalFile.Load for effect {eff}:\n{repr(e)}')



def save_localfile_handler(eff: Effect):
    """
    Example Effect

    {
        'type':     FX.LocalFile.Save,
        'filename': fname,
        'content':  content,
        'settings': settings,
    }
    """
    try:
        content   = eff["content"]
        filename  = eff["filename"]
        settings  = eff.get("settings", {})

        if isinstance(content, Image):
            format = f'.{content.format}'
            if format not in {".svg", ".png", ".jpg", ".jpeg"}: return Error.ValueError(f'Image format needs to be one of these types {{".svg", ".png", ".jpg", ".jpeg"}} got {format} instead.')        
            filename = filename if format == (filename[-4:] or filename[-5:])  else filename + format
            import zstd
            with open(filename, 'wb') as f: f.write(zstd.decompress(content.buffer))
            
        elif "." in filename:
            format = filename[filename.rindex(".") + 1:]
            if format in {"yaml", "yml"}:
                with open(filename, 'w') as f: f.write(yaml.safe_dump(content))
            elif format == "toml":
                with open(filename, 'w') as f: f.write(toml.dumps(content))
            elif format == "csv":
                with open(filename, 'w') as f: f.write(content.to_csv(**settings))
            elif format == "json":
                with open(filename, 'w') as f: f.write(json.dumps(content))
            else:
                return Error.NotImplementedError(f'Unsupported writing to file type {format}!')
        else:
            filename += ".txt"
            with open(filename, 'w') as f: f.write(content)

        return {"filename": filename}
    except Exception as e:
        return Error(f'executing FX.LocalFile.Save for effect {eff}: {repr(e)}')


def write_localfile_handler(eff: Effect):
    """
    Example Effect

    {
        'type':     FX.LocalFile.Write,
        'filename': fname,
        'content':  content,
    }
    """
    content   = eff["content"]
    filename  = eff["filename"]

    mode = "wb"  if isinstance(content, bytes) else "w"
    with open(filename, mode) as f: f.write(content)
    return {"filename": filename}



def system_open_with_handler(eff: Effect):
    """
    This will open the file using the system's default handler application.

    Example
    =======

    {
            'type': FX.LocalFile.SystemOpenWith,
            'filepath': fname,
    }
    """
    try:
        import subprocess, os, platform
        if platform.system() == 'Darwin':       # macOS
            subprocess.Popen(('open', eff['filepath']))
        elif platform.system() == 'Windows':    # Windows
            os.startfile(eff['filepath'])
        else:                                   # linux variants
            subprocess.Popen(('xdg-open', eff['filepath']))

    except Exception as e:
        return Error(f'executing FX.LocalFile.SystemOpenWith for effect {eff}:\n{repr(e)}')



from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
class ZefEventHandler(FileSystemEventHandler):
    def __init__(self, created = None, modified = None, moved = None, deleted = None):
        super().__init__()
        self.created = created
        self.modified = modified
        self.moved = moved
        self.deleted = deleted
    def on_created(self, event):
        super().on_created(event)
        if self.created: self.created(event)
        
    def on_modified(self, event):
        super().on_modified(event)
        if self.modified: self.modified(event)
    def on_moved(self, event):
        super().on_moved(event)
        if self.moved: self.moved(event)
    
    def on_deleted(self, event):
        super().on_deleted(event)
        if self.deleted: self.deleted(event)

def monitor_path_handler(eff: Effect):
    """
    Watches for changes in files and nested directories at given path for changes.
    These changes include, file creation, modification, deletion and moving.

    If a change handler is set, once an event of the type is triggered the handler is called.

    Example
    =======

    {
            'type': FX.LocalFile.MonitorPath,
            'path': fpath,
            'recursive': True,           # default is False
            'created_handler': f1,       # default None
            'modified_handler': f2,      # default None
            'moved_handler': f3,         # default None
            'deleted_handler': f4,       # default None
    }
    """
    assert 'path' in eff, "path is required for FX.LocalFile.MonitorPath"
    path = eff['path']
    recursive = eff.get('recursive', False)

    created_handler = eff.get("created_handler", None)
    modified_handler = eff.get("modified_handler", None)
    moved_handler = eff.get("moved_handler", None)
    deleted_handler = eff.get("deleted_handler", None)

    event_handler = ZefEventHandler(created = created_handler, modified = modified_handler, moved = moved_handler, deleted = deleted_handler)

    observer = Observer()
    observer.schedule(event_handler, path, recursive=recursive)
    observer.start()

    return {"observer_object": observer}


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import *
from ..core import *
from ..ops import *
from ..core.zef_functions import func

@func
def is_a_component(component, vt):
    return isinstance(component, ValueType_) and without_absorbed(component) == vt

#--------------------------TEXT--------------------------------------
def dispatch_rich_text(component):
    import rich.text as rt

    def resolve_style(d):
        if "style" in d: return dispatch_rich_style(d["style"])
        return dispatch_rich_style(Style(**d))

    def resolve_attributes(d):
        allowed_keys = ["justify", "overflow", "no_wrap", "tab_size"]
        attributes = select_keys(d, *allowed_keys)
        return attributes

    def dispatch_on_data_type(data, style):
        # Just a str
        if isinstance(data, str):
            pairs_or_text.append((data, style))
        # Nested Text component
        elif is_a_component(data, Text) : 
            pairs_or_text.extend(dispatch_rich_text(data))
        # List of the above 2
        elif isinstance(data, list):
            [dispatch_on_data_type(x, style) for x in data]
        else:
            raise ValueError("Text's data field could contain only a str, Text component, or a list composed of the two.")
    
    pairs_or_text = []
    internals = component | absorbed | collect
    assert isinstance(internals[0], dict), "First absorbed argument for ZefUI Text should be of type dict!"
    data = internals[0].get("data", "")
    attributes = resolve_attributes(internals[0])
    style      = resolve_style(internals[0])
    dispatch_on_data_type(data, style)

    return rt.Text.assemble(*pairs_or_text, **attributes)



#--------------------------Code--------------------------------------
def dispatch_rich_syntax(component):
    import rich.syntax as rs

    def resolve_attributes(d):
        allowed_keys = ["line_numbers", "theme", "tab_size", "indent_guides"]
        attributes = select_keys(d, *allowed_keys)
        attributes["lexer"] = d.get("language", "python")
        return attributes
    
    internals = component | absorbed | collect
    assert isinstance(internals[0], dict), "First absorbed argument for ZefUI Code should be of type dict!"
    data = internals[0].get("data", "")
    attributes = resolve_attributes(internals[0])
    return rs.Syntax(data, **attributes)


#--------------------------Style--------------------------------------
def dispatch_rich_style(component):
    def resolve_styles(d):
        from rich.style import Style
        allowed_keys = ["bold", "italic", "strike", "underline", "overline", "color"]
        styles = select_keys(d, *allowed_keys)
        styles["bgcolor"] = d.get("background_color", None)
        return Style(**styles)
    
    internals = component | absorbed | collect
    assert isinstance(internals[0], dict), "First absorbed argument for ZefUI Style should be of type dict!"
    return resolve_styles(internals[0])



#--------------------------Table--------------------------------------
def dispatch_rich_table(component):
    import rich.table as rt

    def handle_nested_components(maybe_component):
        if isinstance(maybe_component, str):
            return maybe_component
        elif is_a_component(maybe_component, Text):
            return dispatch_rich_text(maybe_component)
        elif is_a_component(maybe_component, Column):
            return dispatch_rich_column(maybe_component)
        elif is_a_component(maybe_component, Style):
            return dispatch_rich_style(maybe_component)
        elif isinstance(maybe_component, tuple):
            return tuple([handle_nested_components(el) for el in maybe_component])
        else:
            raise NotImplementedError(f"{maybe_component}:{type(maybe_component)}")

    def resolve_attributes(d):
        allowed_keys = ["row_styles", "title", "box", "expand", "show_header", "padding", "show_footer", "show_edge", "show_lines", "width"]
        attributes       = select_keys(d, *allowed_keys)
        # row_styles: List of attributes or strings
        if "row_styles" in attributes:
            attributes["row_styles"] = [handle_nested_components(x) for x in attributes["row_styles"]]
        # title: String or Text
        if "title" in attributes:
            attributes["title"] = handle_nested_components(attributes["title"])
        # box: str -> box.Attribute
        if "box" in attributes:
            attributes["box"] = box_constants_mapping(attributes["box"])
        return attributes
    
    def resolve_row_and_cols(rows, cols):
        return [handle_nested_components(x) for x in rows], [handle_nested_components(x) for x in cols]

    internals = component | absorbed | collect
    assert isinstance(internals[0], dict), "First absorbed argument for ZefUI Table should be of type dict!"
    rows, cols = resolve_row_and_cols(internals[0].get('rows', []), internals[0].get('cols', []))
    attributes = resolve_attributes(internals[0])

    rich_table = rt.Table(*cols, **attributes)
    [rich_table.add_row(*row) for row in rows]
    return rich_table

#--------------------------Column--------------------------------------
def dispatch_rich_column(component):
    import rich.table as rt

    def resolve_attributes(d):
        allowed_keys = ["header_style", "footer_style", "style", "justify", "vertical", "width", "min_width", "max_width", "ratio", "no_wrap"]
        attributes = select_keys(d, *allowed_keys)
        # Resolve the non-string styles if found
        for special_key in ["header_style", "footer_style", "style"]:
            if special_key in attributes and is_a_component(attributes[special_key], Style):
                attributes[special_key] = dispatch_rich_style(attributes[special_key])
        
        return attributes
    
    internals = component | absorbed | collect
    assert isinstance(internals[0], dict), "First absorbed argument for ZefUI Column should be of type dict!"
    
    data = internals[0].get("data", "")
    if is_a_component(data, Text): data = dispatch_rich_text(data)
    attributes = resolve_attributes(internals[0])

    return rt.Column(data, **attributes)


#--------------------------Frame--------------------------------------
def dispatch_rich_panel(component):
    import rich.panel as rp

    def resolve_attributes(d):
        allowed_keys = ["title", "subtitle", "box"]
        attributes = select_keys(d, *allowed_keys)
        if "title" in attributes and is_a_component(attributes["title"], Text): 
            attributes["title"] = dispatch_rich_text(attributes["title"])
        
        if "subtitle" in attributes and is_a_component(attributes["subtitle"], Text): 
            attributes["subtitle"] = dispatch_rich_text(attributes["subtitle"])

        if "box" in attributes:
            attributes["box"] = box_constants_mapping(attributes["box"])

        return attributes
    
    internals = component | absorbed | collect
    assert isinstance(internals[0], dict), "First absorbed argument for ZefUI Code should be of type dict!"
    if "data" not in internals[0]: raise ValueError("Can't render Frame without any data to display!")
    data = internals[0].get("data")
    data = match_and_dispatch(data)
    attributes = resolve_attributes(internals[0])
    return rp.Panel(data, **attributes)


#--------------------------HStack,VStack--------------------------------------
def dispatch_rich_stack(component):
    import rich.table as rt

    def handle_nested_components(maybe_component):
        if isinstance(maybe_component, str):
            return maybe_component
        else:
            return match_and_dispatch(maybe_component)

    def resolve_attributes(d):
        allowed_keys = ["expand", "padding", "pad_edge"]
        attributes = select_keys(d, *allowed_keys)
        return attributes
    
    stack_type = str(component | without_absorbed | collect)
    internals = component | absorbed | collect
    assert isinstance(internals[0], dict), "First absorbed argument for ZefUI Table should be of type dict!"
    data = internals[0].get('data', [])
    assert isinstance(data, list), f"Data for ZefUI Stack should be of type list! {data} was passed"
    cols = internals[0].get('cols', [])
    assert isinstance(cols, list), f"Data for ZefUI Stack should be of type list! {cols} was passed"
    data = [handle_nested_components(c) for c in data]
    cols = [handle_nested_components(c) for c in cols]
    attributes = resolve_attributes(internals[0])

    rich_grid = rt.Table.grid(*cols, **attributes)
    if stack_type == "HStack":
        rich_grid.add_row(*data)
    elif stack_type == "VStack":
        [rich_grid.add_row(row) for row in data]
    return rich_grid

#--------------------------Bullet/Numbered List--------------------------------------
def dispatch_bullet_or_numbered_list(component):
        list_type = str(component | without_absorbed | collect)
        if list_type == 'BulletList':
            dispatch_type = lambda _: '- ' 
        else:
            dispatch_type = lambda i: str(i + 1) +'. '
        
        internals = component | absorbed | collect
        assert isinstance(internals[0], dict), f"First absorbed argument for ZefUI {list_type} should be of type dict!"
        data  = internals[0].get('data', [])
        
        data_modified = []
        for i, el in enumerate(data):
            if isinstance(el, str): 
                data_modified.append(dispatch_type(i) + el)
            else:
                data_modified.append(HStack(data=[dispatch_type(i), el]))

        heading = internals[0].get('heading', "")
        if heading: data_modified = data_modified | prepend[heading] | collect

        allowed_keys = ["padding", "expand"]
        attributes = select_keys(internals[0], *allowed_keys)

        return dispatch_rich_stack(VStack(data_modified, **attributes))

#--------------------------Markdown--------------------------------------
def dispatch_rich_markdown(component):
    import rich.markdown as rm

    def resolve_style(d):
        if "style" in d: return dispatch_rich_style(d["style"])
        return dispatch_rich_style(Style(**d))
    
    internals = component | absorbed | collect
    assert isinstance(internals[0], dict), "First absorbed argument for ZefUI Table should be of type dict!"
    style = resolve_style(internals[0])

    return rm.Markdown(component_to_markdown_string(component), style = style)


def component_to_markdown_string(component):
    
    def dispatch_bullet_or_numbered_list(component):
        list_type = str(component | without_absorbed | collect)
        if list_type == 'BulletList':
            dispatch_type = lambda i: '- ' 
        else:
            dispatch_type = lambda i: str(i + 1) +'. '
        nl = '\n'
        
        internals = component | absorbed | collect
        assert isinstance(internals[0], dict), f"First absorbed argument for ZefUI {list_type} should be of type dict!"
        data  = internals[0].get('data', [])
        heading = internals[0].get('heading', "")
        heading_level = internals[0].get('heading_level', 3)
        
        if heading: heading = f"{'#' * heading_level} {heading}"
        
        return f"""
{heading}
{nl.join([f'{dispatch_type(i)}{e}' for i,e in enumerate(data)])}
    """

    def dispatch_paragraph(component):
        internals = component | absorbed | collect
        assert isinstance(internals[0], dict), "First absorbed argument for ZefUI Paragraph should be of type dict!"
        data  = internals[0].get('data', "")
        heading = internals[0].get('heading', "")
        heading_level = internals[0].get('heading_level', 2)
        if heading: heading = f"{'#' * heading_level} {heading}"

    
        return f"""
{heading}
{data}
    """

    return component | match[
        (Is[is_a_component[BulletList]], dispatch_bullet_or_numbered_list),
        (Is[is_a_component[NumberedList]], dispatch_bullet_or_numbered_list),
        (Is[is_a_component[Paragraph]], dispatch_paragraph),
    ] | collect


#-------------------Dispatch-------------------------------------
def box_constants_mapping(box_style: str):
    from rich import box
    str_to_constant = {
        'ascii':                   box.ASCII,
        'square':                  box.SQUARE,                
        'minimal':                 box.MINIMAL,  
        'minimal_heavy_head':      box.MINIMAL_HEAVY_HEAD,     
        'minimal_double_head':     box.MINIMAL_DOUBLE_HEAD,           
        'simple':                  box.SIMPLE,    
        'simple_head':             box.SIMPLE_HEAD,    
        'heavy':                   box.HEAVY,                 
        'heavy_edge':              box.HEAVY_EDGE,             
        'heavy_head':              box.HEAVY_HEAD,
        'double':                  box.DOUBLE,               
        'double_edge':             box.DOUBLE_EDGE,
        'simple_heavy':            box.SIMPLE_HEAVY,            
        'horizontals':             box.HORIZONTALS,              
        'rounded':                 box.ROUNDED,
    }
    return str_to_constant.get(box_style, box.ROUNDED)

def match_and_dispatch(component):
    return component | match[
        (Is[is_a_component[Text]], dispatch_rich_text),
        (Is[is_a_component[Code]], dispatch_rich_syntax),
        (Is[is_a_component[Style]], dispatch_rich_style),
        (Is[is_a_component[Table]], dispatch_rich_table),
        (Is[is_a_component[Column]], dispatch_rich_column),
        (Is[is_a_component[Frame]], dispatch_rich_panel),
        (Is[is_a_component[HStack]], dispatch_rich_stack),
        (Is[is_a_component[VStack]], dispatch_rich_stack),
        (Is[is_a_component[Paragraph]], dispatch_rich_markdown),
        (Is[is_a_component[BulletList]], dispatch_bullet_or_numbered_list),
        (Is[is_a_component[NumberedList]], dispatch_bullet_or_numbered_list),
    ] | collect

def print_rich(displayable):
    import rich
    console = rich.console.Console()
    displayable = match_and_dispatch(displayable)
    console.print(displayable)

show = run[print_rich]
displayable = run[match_and_dispatch]


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from icecream import ic
import asyncio
from functools import partial as P

from zefdb import *
from zefdb.zefops import *
from zefdb import zefui

import socketio
sio = socketio.AsyncServer(logger=True, engineio_logger=True, cors_allowed_origins="*")

import sanic
sanic_app = sanic.Sanic("zefui")
sio.attach(sanic_app)


import logging
logging.getLogger('sanic_cors').level = logging.DEBUG

def start_server():
    asyncio.get_event_loop().set_debug(True)
    sanic_app.run(port=3001)



from zefdb.zefui.server_temp import *


####

make_app_funcs = {}

template_graph = generate_template_graph_programmatically()


####

@sio.event
def connect(sid, environ, auth):
    print("connecting ", sid)

    # TODO: If this is a server triggered from the REPL, send the current state immediately

@sio.event
async def page_request(sid, data):
    # TODO: If this is a server triggered from the REPL, ignore this request.

    desired_app = data["pageId"]
    make_app = make_app_funcs[desired_app]

    # Would hopefully have the ability to add stuff into data later and pass this in.
    root_component = make_app()

    await sio.emit("page_config", {
        "layout": ["asdf"],
        "layoutElements": {},
        "components": {},
        "state": {},
    })

    make_component(root_component, sid)

    outgoing_stream,incoming_stream = create_zefui_streams(Graph(root_component) | uid)

    async def send_updates(updates):
        print("In send updates")
        print(updates)

        # TODO: After the frontend accepts lists can change this
        # await sio.emit("state_update", updates)
        for update in updates:
            await sio.emit("update", update)
    asyncio.create_task(sync_map_rx_stream(send_updates, outgoing_stream))
    # outgoing_stream.subscribe(on_next=lambda x: print("Got an update", x))

    # Just hacking for now
    await asyncio.sleep(1)
    await send_updates([{#"layout": ["asdf"],
                         "layoutElements": {"asdf":
                                            {"<>": "div",
                                             "children":
                                             {"__set": [f"COMPONENT({root_component|uid})"]
                                              }
                                              }}}])

    await asyncio.sleep(1)
    root_component | attach[RT.Placeholder, "asdf"]
    await asyncio.sleep(3)
    terminate(root_component >> RT.Value)
    await asyncio.sleep(3)



    await sio.save_session(sid, {"root_component": root_component,
                                 "zefui_output_stream": outgoing_stream,
                                 "zefui_input_stream": incoming_stream})


#######

@sio.event
async def value_change(sid, data):
    session = await sio.get_session(sid)
    root_component = session["root_component"]

    # TODO: This is a just a quick test version. In the future, dispatch to
    # appropriate component logic.
    z = g | instances[now][ET.Input] | only
    ae = z >> RT.Value
    ae <= "triggered"
    print("value_change")


#######

def make_app_basic():
    ui_model = Graph()
    I = lambda *args: instantiate(*args, ui_model)

    z_root = I(ET.Input) | attach[RT.Value, "asdf"]
    
    global stored_ui_model
    stored_ui_model = ui_model
    
    return z_root|now
make_app_funcs["master-schedule"] = make_app_basic


if __name__ == "__main__":
    start_server()




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
backend = os.environ.get("SQL_UI_BACKEND", None)
if backend is None:
    try:
        import sdl2
        backend = "sdl2"
    except ImportError:
        pass
# The pygame backend seems broken
# if backend is None:
#     try:
#         import pygame
#         backend = "pygame"
#     except ImportError:
#         pass
if backend is None:
    try:
        import glfw
        backend = "glfw"
    except ImportError:
        pass

if backend is None:
    raise ImportError("""Can't import either sdl2 or glfw for imgui backend support in sql_ui wizard.

Please install either of these backends using
`pip install 'imgui[sdl2]'`
or
`pip install 'imgui[glfw]'`

You may have to install the corresponding libraries on your system as well, e.g.
`brew install SDL2`.
""")

import imgui
from . import ui 

if __name__ == "__main__":
    import sys
    assert len(sys.argv) == 2, "Need one argument of the yaml file to edit."
    filepath = sys.argv[1]


    if backend == "sdl2":
        from .wizard_sdl2 import main
    elif backend == "pygame":
        from .wizard_pygame import main
    elif backend == "glfw":
        from .wizard_glfw import main
    main("SQL Import Wizard", ui.init_state, ui.render, filepath)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *

class MyTestCase(unittest.TestCase):
    def test_zefrefs_from_slices(self):

        g = Graph()
        z = AET.String | g | run
        z | assign["a"] | g | run
        z | assign["b"] | g | run
        z | terminate | g | run

        g2 = Graph()
        z2 = z | g2 | run

        ctx = z | instantiation_tx | collect
        # Note: slice 1 is created on Graph(), so this is slice 2
        self.assertEqual(ctx | to_graph_slice | graph_slice_index | collect, 2)

        self.assertEqual(ctx | events[Instantiated] | length | collect, 2)
        self.assertEqual(ctx | events[Assigned] | length | collect, 0)
        self.assertEqual(ctx | events[events[Terminated]] | length | collect, 0)

        txs = z | now[allow_tombstone] | preceding_events[Assigned] | map[absorbed | first | frame | to_tx] | collect
        self.assertEqual([ctx  | frame | graph_slice_index | collect for ctx in txs], [3, 4])
        for ctx in txs:
            self.assertEqual(ctx | events[Instantiated] | length | collect, 1)
            self.assertEqual(ctx | events[Assigned] | length | collect, 1)
            self.assertEqual(ctx | events[Terminated] | length | collect, 0)
            

        ctx = z | now[allow_tombstone] | termination_tx | collect
        self.assertEqual(ctx | to_graph_slice | graph_slice_index | collect, 5)

        self.assertEqual(ctx | events[Instantiated] | length | collect, 0)
        self.assertEqual(ctx | events[Assigned] | length | collect, 0)
        self.assertEqual(ctx | events[Terminated] | length | collect, 1)

        ctx = z2 | termination_tx | collect
        self.assertEqual(ctx, g2[42])


if __name__ == '__main__':
    unittest.main()
